@@ -26,17 +26,19 @@ extern "C" {
 #include <osmocom/core/linuxlist.h>
 #include <osmocom/core/rate_ctr.h>
 #include <osmocom/core/stat_item.h>
-#include <osmocom/core/timer.h>
-#include <osmocom/core/gsmtap.h>
 #include <osmocom/gsm/l1sap.h>
 #include <osmocom/gsm/protocol/gsm_04_08.h>
+	#include <mslot_class.h>
 }
 
+#include <gsm_rlcmac.h>
 #include "poll_controller.h"
 #include "sba.h"
 #include "tbf.h"
 #include "gprs_ms_storage.h"
 #include "gprs_coding_scheme.h"
+#include <cxx_linuxlist.h>
+#include <pdch.h>
 #endif
 
 #include <stdint.h>
@@ -64,86 +66,6 @@ enum pcu_gsmtap_category {
 struct BTS;
 struct GprsMs;
 
-/*
- * PDCH instance
- */
-struct gprs_rlcmac_pdch {
-#ifdef __cplusplus
-	struct gprs_rlcmac_paging *dequeue_paging();
-	struct msgb *packet_paging_request();
-
-	void add_paging(struct gprs_rlcmac_paging *pag);
-
-	void free_resources();
-
-	bool is_enabled() const;
-
-	void enable();
-	void disable();
-
-	/* dispatching of messages */
-	int rcv_block(uint8_t *data, uint8_t len, uint32_t fn,
-		struct pcu_l1_meas *meas);
-	int rcv_block_gprs(uint8_t *data, uint8_t data_len, uint32_t fn,
-		struct pcu_l1_meas *meas, GprsCodingScheme cs);
-	int rcv_data_block(uint8_t *data, uint8_t data_len, uint32_t fn,
-		struct pcu_l1_meas *meas, GprsCodingScheme cs);
-
-	gprs_rlcmac_bts *bts_data() const;
-	BTS *bts() const;
-	uint8_t trx_no() const;
-
-	struct gprs_rlcmac_ul_tbf *ul_tbf_by_tfi(uint8_t tfi);
-	struct gprs_rlcmac_dl_tbf *dl_tbf_by_tfi(uint8_t tfi);
-
-	void attach_tbf(gprs_rlcmac_tbf *tbf);
-	void detach_tbf(gprs_rlcmac_tbf *tbf);
-
-	unsigned num_tbfs(enum gprs_rlcmac_tbf_direction dir) const;
-
-	void reserve(enum gprs_rlcmac_tbf_direction dir);
-	void unreserve(enum gprs_rlcmac_tbf_direction dir);
-	unsigned num_reserved(enum gprs_rlcmac_tbf_direction dir) const;
-
-	uint8_t assigned_usf() const;
-	uint32_t assigned_tfi(enum gprs_rlcmac_tbf_direction dir) const;
-#endif
-
-	uint8_t m_is_enabled; /* TS is enabled */
-	uint8_t tsc; /* TSC of this slot */
-	uint8_t next_ul_tfi; /* next uplink TBF/TFI to schedule (0..31) */
-	uint8_t next_dl_tfi; /* next downlink TBF/TFI to schedule (0..31) */
-	uint8_t next_ctrl_prio; /* next kind of ctrl message to schedule */
-	struct llist_head paging_list; /* list of paging messages */
-	uint32_t last_rts_fn; /* store last frame number of RTS */
-
-	/* back pointers */
-	struct gprs_rlcmac_trx *trx;
-	uint8_t ts_no;
-
-#ifdef __cplusplus
-private:
-	int rcv_control_block(const uint8_t *data, uint8_t data_len, bitvec *rlc_block, uint32_t fn);
-
-	void rcv_control_ack(Packet_Control_Acknowledgement_t *, uint32_t fn);
-	void rcv_control_dl_ack_nack(Packet_Downlink_Ack_Nack_t *, uint32_t fn);
-	void rcv_control_egprs_dl_ack_nack(EGPRS_PD_AckNack_t *, uint32_t fn);
-	void rcv_resource_request(Packet_Resource_Request_t *t, uint32_t fn);
-	void rcv_measurement_report(Packet_Measurement_Report_t *t, uint32_t fn);
-	gprs_rlcmac_tbf *tbf_from_list_by_tfi(
-		LListHead<gprs_rlcmac_tbf> *tbf_list, uint8_t tfi,
-		enum gprs_rlcmac_tbf_direction dir);
-	gprs_rlcmac_tbf *tbf_by_tfi(uint8_t tfi,
-		enum gprs_rlcmac_tbf_direction dir);
-#endif
-
-	uint8_t m_num_tbfs[2];
-	uint8_t m_num_reserved[2];
-	uint8_t m_assigned_usf; /* bit set */
-	uint32_t m_assigned_tfi[2]; /* bit set */
-	struct gprs_rlcmac_tbf *m_tbfs[2][32];
-};
-
 struct gprs_rlcmac_trx {
 	void *fl1h;
 	uint16_t arfcn;
@@ -162,8 +84,7 @@ struct gprs_rlcmac_trx {
 #ifdef __cplusplus
 extern "C" {
 #endif
-void bts_update_tbf_ta(const char *p, uint32_t fn, uint8_t trx_no, uint8_t ts,
-		       uint8_t ta);
+void bts_update_tbf_ta(const char *p, uint32_t fn, uint8_t trx_no, uint8_t ts, int8_t ta, bool is_rach);
 #ifdef __cplusplus
 }
 #endif
@@ -173,6 +94,7 @@ void bts_update_tbf_ta(const char *p, uint32_t fn, uint8_t trx_no, uint8_t ts,
  * we can start to compile pcu_vty.c with c++ and remove the split.
  */
 struct gprs_rlcmac_bts {
+	bool active;
 	uint8_t bsic;
 	uint8_t fc_interval;
 	uint16_t fc_bucket_time;
@@ -204,11 +126,9 @@ struct gprs_rlcmac_bts {
 	struct gsmtap_inst *gsmtap;
 	uint32_t gsmtap_categ_mask;
 	struct gprs_rlcmac_trx trx[8];
-	int (*alloc_algorithm)(struct gprs_rlcmac_bts *bts,
-		struct GprsMs *ms,
-		struct gprs_rlcmac_tbf *tbf, uint32_t cust, uint8_t single,
-		int use_tbf);
-	uint32_t alloc_algorithm_curst; /* options to customize algorithm */
+	int (*alloc_algorithm)(struct gprs_rlcmac_bts *bts, struct GprsMs *ms, struct gprs_rlcmac_tbf *tbf,
+			       bool single, int8_t use_tbf);
+
 	uint8_t force_two_phase;
 	uint8_t alpha, gamma;
 	uint8_t egprs_enabled;
@@ -366,14 +286,12 @@ public:
 	gprs_rlcmac_dl_tbf *dl_tbf_by_tfi(uint8_t tfi, uint8_t trx, uint8_t ts);
 	gprs_rlcmac_ul_tbf *ul_tbf_by_tfi(uint8_t tfi, uint8_t trx, uint8_t ts);
 
-	int tfi_find_free(enum gprs_rlcmac_tbf_direction dir, uint8_t *_trx, int8_t use_trx);
+	int tfi_find_free(enum gprs_rlcmac_tbf_direction dir, uint8_t *_trx, int8_t use_trx) const;
 
 	int rcv_imm_ass_cnf(const uint8_t *data, uint32_t fn);
-	uint8_t is_single_block(uint16_t ra, enum ph_burst_type burst_type,
-		uint8_t is_11bit, uint16_t *ms_class, uint16_t *priority);
 
 	uint32_t rfn_to_fn(int32_t rfn);
-	int rcv_rach(uint16_t ra, uint32_t Fn, int16_t qta, uint8_t is_11bit,
+	int rcv_rach(uint16_t ra, uint32_t Fn, int16_t qta, bool is_11bit,
 		enum ph_burst_type burst_type);
 
 	void snd_dl_ass(gprs_rlcmac_tbf *tbf, uint8_t poll, const char *imsi);
@@ -543,33 +461,6 @@ inline LListHead<gprs_rlcmac_tbf>& BTS::dl_tbfs()
 	return m_dl_tbfs;
 }
 
-inline BTS *gprs_rlcmac_pdch::bts() const
-{
-	return trx->bts;
-}
-
-inline unsigned gprs_rlcmac_pdch::num_tbfs(enum gprs_rlcmac_tbf_direction dir) const
-{
-	return m_num_tbfs[dir];
-}
-
-inline unsigned gprs_rlcmac_pdch::num_reserved(
-	enum gprs_rlcmac_tbf_direction dir) const
-{
-	return gprs_rlcmac_pdch::m_num_reserved[dir];
-}
-
-inline uint8_t gprs_rlcmac_pdch::assigned_usf() const
-{
-	return m_assigned_usf;
-}
-
-inline uint32_t gprs_rlcmac_pdch::assigned_tfi(
-	enum gprs_rlcmac_tbf_direction dir) const
-{
-	return m_assigned_tfi[dir];
-}
-
 inline struct rate_ctr_group *BTS::rate_counters() const
 {
 	return m_ratectrs;
@@ -690,15 +581,6 @@ CREATE_STAT_INLINE(ms_present, ms_present_get, STAT_MS_PRESENT);
 
 #undef CREATE_STAT_INLINE
 
-inline gprs_rlcmac_bts *gprs_rlcmac_pdch::bts_data() const
-{
-	return trx->bts->bts_data();
-}
-
-inline uint8_t gprs_rlcmac_pdch::trx_no() const
-{
-	return trx->trx_no;
-}
 #endif
 
 #ifdef __cplusplus
@@ -710,8 +592,4 @@ extern "C" {
 #ifdef __cplusplus
 }
 
-inline bool gprs_rlcmac_pdch::is_enabled() const
-{
-	return m_is_enabled;
-}
 #endif
