@@ -19,11 +19,9 @@
 
 #pragma once
 
-#include <stdio.h>
 #ifdef __cplusplus
 extern "C" {
 #endif
-#include <osmocom/core/linuxlist.h>
 #include <osmocom/core/logging.h>
 #ifdef __cplusplus
 };
@@ -39,6 +37,9 @@ enum {
 	DRLCMACUL,
 	DRLCMACSCHED,
 	DRLCMACMEAS,
+	DTBF,
+	DTBFDL,
+	DTBFUL,
 	DNS,
 	DBSSGP,
 	DPCU,
