@@ -22,9 +22,17 @@
 #include <pcu_l1_if.h>
 #include <bts.h>
 #include <tbf.h>
-
+#include <gprs_debug.h>
+#include <gprs_ms.h>
+#include <rlc.h>
+#include <sba.h>
+#include <pdch.h>
 #include "pcu_utils.h"
 
+extern "C" {
+	#include <osmocom/core/gsmtap.h>
+}
+
 static uint32_t sched_poll(BTS *bts,
 		    uint8_t trx, uint8_t ts, uint32_t fn, uint8_t block_nr,
 		    struct gprs_rlcmac_tbf **poll_tbf,
@@ -49,16 +57,14 @@ static uint32_t sched_poll(BTS *bts,
 		if (ul_tbf->trx->trx_no != trx || !ul_tbf->is_control_ts(ts))
 			continue;
 		/* polling for next uplink block */
-		if (ul_tbf->poll_state == GPRS_RLCMAC_POLL_SCHED
-		 && ul_tbf->poll_fn == poll_fn)
+		if (ul_tbf->poll_scheduled() && ul_tbf->poll_fn == poll_fn)
 			*poll_tbf = ul_tbf;
-		if (ul_tbf->ul_ack_state == GPRS_RLCMAC_UL_ACK_SEND_ACK)
+		if (ul_tbf->ul_ack_state_is(GPRS_RLCMAC_UL_ACK_SEND_ACK))
 			*ul_ack_tbf = ul_tbf;
-		if (ul_tbf->dl_ass_state == GPRS_RLCMAC_DL_ASS_SEND_ASS)
+		if (ul_tbf->dl_ass_state_is(GPRS_RLCMAC_DL_ASS_SEND_ASS))
 			*dl_ass_tbf = ul_tbf;
-		if (ul_tbf->ul_ass_state == GPRS_RLCMAC_UL_ASS_SEND_ASS
-			|| ul_tbf->ul_ass_state ==
-				GPRS_RLCMAC_UL_ASS_SEND_ASS_REJ)
+		if (ul_tbf->ul_ass_state_is(GPRS_RLCMAC_UL_ASS_SEND_ASS)
+		    || ul_tbf->ul_ass_state_is(GPRS_RLCMAC_UL_ASS_SEND_ASS_REJ))
 			*ul_ass_tbf = ul_tbf;
 /* FIXME: Is this supposed to be fair? The last TBF for each wins? Maybe use llist_add_tail and skip once we have all
 states? */
@@ -70,13 +76,12 @@ states? */
 		if (dl_tbf->trx->trx_no != trx || !dl_tbf->is_control_ts(ts))
 			continue;
 		/* polling for next uplink block */
-		if (dl_tbf->poll_state == GPRS_RLCMAC_POLL_SCHED
-		 && dl_tbf->poll_fn == poll_fn)
+		if (dl_tbf->poll_scheduled() && dl_tbf->poll_fn == poll_fn)
 			*poll_tbf = dl_tbf;
-		if (dl_tbf->dl_ass_state == GPRS_RLCMAC_DL_ASS_SEND_ASS)
+		if (dl_tbf->dl_ass_state_is(GPRS_RLCMAC_DL_ASS_SEND_ASS))
 			*dl_ass_tbf = dl_tbf;
-		if (dl_tbf->ul_ass_state == GPRS_RLCMAC_UL_ASS_SEND_ASS
-		 || dl_tbf->ul_ass_state == GPRS_RLCMAC_UL_ASS_SEND_ASS_REJ)
+		if (dl_tbf->ul_ass_state_is(GPRS_RLCMAC_UL_ASS_SEND_ASS)
+		    || dl_tbf->ul_ass_state_is(GPRS_RLCMAC_UL_ASS_SEND_ASS_REJ))
 			*ul_ass_tbf = dl_tbf;
 	}
 
@@ -139,13 +144,11 @@ static struct msgb *sched_select_ctrl_msg(
 		 * because they may kill the TBF when the CONTROL ACK is
 		 * received, thus preventing the others from being processed.
 		 */
-		if (tbf == ul_ass_tbf && tbf->ul_ass_state ==
-				GPRS_RLCMAC_UL_ASS_SEND_ASS_REJ)
+		if (tbf == ul_ass_tbf && tbf->ul_ass_state_is(GPRS_RLCMAC_UL_ASS_SEND_ASS_REJ))
 			msg = ul_ass_tbf->create_packet_access_reject();
 		else if (tbf == ul_ass_tbf && tbf->direction ==
 				GPRS_RLCMAC_DL_TBF)
-			if (tbf->ul_ass_state ==
-					GPRS_RLCMAC_UL_ASS_SEND_ASS_REJ)
+			if (tbf->ul_ass_state_is(GPRS_RLCMAC_UL_ASS_SEND_ASS_REJ))
 				msg = ul_ass_tbf->create_packet_access_reject();
 			else
 				msg = ul_ass_tbf->create_ul_ass(fn, ts);
@@ -207,27 +210,41 @@ static struct msgb *sched_select_ctrl_msg(
 	return NULL;
 }
 
+static inline enum tbf_dl_prio tbf_compute_priority(const struct gprs_rlcmac_bts *bts, struct gprs_rlcmac_dl_tbf *tbf,
+						    uint8_t ts, uint32_t fn, int age)
+{
+	const gprs_rlc_dl_window *w = tbf->window();
+	int age_thresh1 = msecs_to_frames(200),
+		age_thresh2 = msecs_to_frames(OSMO_MIN(BTS::TIMER_T3190_MSEC/2, bts->dl_tbf_idle_msec));
+
+	if (tbf->is_control_ts(ts) && tbf->need_control_ts())
+		return DL_PRIO_CONTROL;
+
+	if (tbf->is_control_ts(ts) && age > age_thresh2 && age_thresh2 > 0)
+		return DL_PRIO_HIGH_AGE;
+
+	if ((tbf->state_is(GPRS_RLCMAC_FLOW) && tbf->have_data()) || w->resend_needed() >= 0)
+		return DL_PRIO_NEW_DATA;
+
+	if (tbf->is_control_ts(ts) && age > age_thresh1 && tbf->keep_open(fn))
+		return DL_PRIO_LOW_AGE;
+
+	if (!w->window_empty())
+		return DL_PRIO_SENT_DATA;
+
+	return DL_PRIO_NONE;
+}
+
 static struct msgb *sched_select_downlink(struct gprs_rlcmac_bts *bts,
 		    uint8_t trx, uint8_t ts, uint32_t fn,
 		    uint8_t block_nr, struct gprs_rlcmac_pdch *pdch)
 {
 	struct msgb *msg = NULL;
 	struct gprs_rlcmac_dl_tbf *tbf, *prio_tbf = NULL;
-	enum {
-		DL_PRIO_NONE,
-		DL_PRIO_SENT_DATA, /* the data has been sent and not (yet) nacked */
-		DL_PRIO_LOW_AGE,   /* the age has reached the first threshold */
-		DL_PRIO_NEW_DATA,  /* the data has not been sent yet or nacked */
-		DL_PRIO_HIGH_AGE,  /* the age has reached the second threshold */
-		DL_PRIO_CONTROL,   /* a control block needs to be sent */
-	} prio, max_prio = DL_PRIO_NONE;
+	enum tbf_dl_prio prio, max_prio = DL_PRIO_NONE;
 
 	uint8_t i, tfi, prio_tfi;
 	int age;
-	const int age_thresh1 = msecs_to_frames(200);
-	const int high_prio_msecs =
-		OSMO_MIN(BTS::TIMER_T3190_MSEC/2, bts->dl_tbf_idle_msec);
-	const int age_thresh2 = msecs_to_frames(high_prio_msecs);
 
 	/* select downlink resource */
 	for (i = 0, tfi = pdch->next_dl_tfi; i < 32;
@@ -251,20 +268,8 @@ static struct msgb *sched_select_downlink(struct gprs_rlcmac_bts *bts,
 		age = tbf->frames_since_last_poll(fn);
 
 		/* compute priority */
-		if (tbf->is_control_ts(ts) && tbf->need_control_ts())
-			prio = DL_PRIO_CONTROL;
-		else if (tbf->is_control_ts(ts) &&
-			age > age_thresh2 && age_thresh2 > 0)
-			prio = DL_PRIO_HIGH_AGE;
-		else if ((tbf->state_is(GPRS_RLCMAC_FLOW) && tbf->have_data()) ||
-			tbf->m_window.resend_needed() >= 0)
-			prio = DL_PRIO_NEW_DATA;
-		else if (tbf->is_control_ts(ts) &&
-			age > age_thresh1 && tbf->keep_open(fn))
-			prio = DL_PRIO_LOW_AGE;
-		else if (!tbf->m_window.window_empty())
-			prio = DL_PRIO_SENT_DATA;
-		else
+		prio = tbf_compute_priority(bts, tbf, ts, fn, age);
+		if (prio == DL_PRIO_NONE)
 			continue;
 
 		/* get the TBF with the highest priority */
