@@ -15,6 +15,7 @@
 #include <sysmo_l1_if.h>
 #include <gprs_debug.h>
 #include <pcu_l1_if.h>
+#include <pdch.h>
 #include <bts.h>
 
 extern void *tall_pcu_ctx;
@@ -188,7 +189,7 @@ static int handle_ph_data_ind(struct femtol1_hdl *fl1h,
 
 	get_meas(&meas, &data_ind->measParam);
 	bts_update_tbf_ta("PH-DATA", data_ind->u32Fn, fl1h->trx_no,
-			  data_ind->u8Tn, qta2ta(meas.bto));
+			  data_ind->u8Tn, sign_qta2ta(meas.bto), false);
 
 	switch (data_ind->sapi) {
 	case GsmL1_Sapi_Pdtch:
@@ -237,7 +238,7 @@ static int handle_ph_ra_ind(struct femtol1_hdl *fl1h, GsmL1_PhRaInd_t *ra_ind)
 
 	DEBUGP(DL1IF, "Rx PH-RA.ind");
 	bts_update_tbf_ta("PH-RA", ra_ind->u32Fn, fl1h->trx_no, ra_ind->u8Tn,
-			  qta2ta(ra_ind->measParam.i16BurstTiming));
+			qta2ta(ra_ind->measParam.i16BurstTiming), true);
 
 	return 0;
 }
