@@ -29,15 +29,18 @@ extern "C" {
 #include <osmocom/core/talloc.h>
 #include <osmocom/core/select.h>
 #include <osmocom/core/msgb.h>
+#include <osmocom/core/linuxlist.h>
+#include <osmocom/core/logging.h>
+#include <osmocom/core/timer.h>
 }
 
-#include <gprs_rlcmac.h>
 #include <pcu_l1_if.h>
 #include <gprs_debug.h>
 #include <gprs_bssgp_pcu.h>
 #include <osmocom/pcu/pcuif_proto.h>
 #include <bts.h>
 #include <tbf.h>
+#include <pdch.h>
 
 extern void *tall_pcu_ctx;
 
@@ -55,7 +58,23 @@ struct pcu_sock_state {
 	struct llist_head upqueue;	/* queue for sending messages */
 } *pcu_sock_state = NULL;
 
-static void pcu_sock_timeout(void *_priv);
+static void pcu_sock_timeout(void *_priv)
+{
+	pcu_l1if_open();
+}
+
+static void pcu_tx_txt_retry(void *_priv)
+{
+	struct gprs_rlcmac_bts *bts = bts_main_data();
+	struct pcu_sock_state *state = pcu_sock_state;
+
+	if (bts->active)
+		return;
+
+	LOGP(DL1IF, LOGL_INFO, "Sending version %s to BTS.\n", PACKAGE_VERSION);
+	pcu_tx_txt_ind(PCU_VERSION, "%s", PACKAGE_VERSION);
+	osmo_timer_schedule(&state->timer, 5, 0);
+}
 
 int pcu_sock_send(struct msgb *msg)
 {
@@ -263,11 +282,12 @@ int pcu_l1if_open(void)
 	rc = connect(bfd->fd, (struct sockaddr *) &local, namelen);
 	if (rc != 0) {
 		LOGP(DL1IF, LOGL_ERROR, "Failed to connect to the osmo-bts"
-			" PCU socket, delaying... '%s'\n", local.sun_path);
+			" PCU socket (%s), delaying... '%s'\n",
+			strerror(errno), local.sun_path);
 		pcu_sock_state = state;
 		close(bfd->fd);
 		bfd->fd = -1;
-		state->timer.cb = pcu_sock_timeout;
+		osmo_timer_setup(&state->timer, pcu_sock_timeout, NULL);
 		osmo_timer_schedule(&state->timer, 5, 0);
 		return 0;
 	}
@@ -292,6 +312,10 @@ int pcu_l1if_open(void)
 	LOGP(DL1IF, LOGL_INFO, "Sending version %s to BTS.\n", PACKAGE_VERSION);
 	pcu_tx_txt_ind(PCU_VERSION, "%s", PACKAGE_VERSION);
 
+	/* Schedule a timer so we keep trying until the BTS becomes active. */
+	osmo_timer_setup(&state->timer, pcu_tx_txt_retry, NULL);
+	osmo_timer_schedule(&state->timer, 5, 0);
+
 	return 0;
 }
 
@@ -311,8 +335,3 @@ void pcu_l1if_close(void)
 	talloc_free(state);
 	pcu_sock_state = NULL;
 }
-
-static void pcu_sock_timeout(void *_priv)
-{
-	pcu_l1if_open();
-}
