@@ -17,22 +17,34 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  */
 
-#include <gprs_bssgp_pcu.h>
-#include <arpa/inet.h>
 #include <pcu_l1_if.h>
 #include <gprs_rlcmac.h>
 #include <gsm_timer.h>
 #include <gprs_debug.h>
 #include <unistd.h>
+#include <stdbool.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include <errno.h>
 #include <getopt.h>
 #include <signal.h>
 #include <sched.h>
 #include <bts.h>
+#include <gprs_coding_scheme.h>
+#include <osmocom/pcu/pcuif_proto.h>
 extern "C" {
 #include "pcu_vty.h"
+#include <osmocom/gprs/gprs_bssgp.h>
+#include <osmocom/gprs/gprs_ns.h>
 #include <osmocom/vty/telnet_interface.h>
-#include <osmocom/vty/logging.h>
+#include <osmocom/vty/command.h>
+#include <osmocom/vty/vty.h>
 #include <osmocom/vty/ports.h>
+#include <osmocom/core/rate_ctr.h>
+#include <osmocom/core/select.h>
+#include <osmocom/core/application.h>
+#include <osmocom/core/msgb.h>
 #include <osmocom/core/stats.h>
 #include <osmocom/core/gsmtap.h>
 #include <osmocom/core/gsmtap_util.h>
@@ -40,10 +52,11 @@ extern "C" {
 
 extern struct gprs_nsvc *nsvc;
 uint16_t spoof_mcc = 0, spoof_mnc = 0;
+bool spoof_mnc_3_digits = false;
 static int config_given = 0;
 static char *config_file = strdup("osmo-pcu.cfg");
 extern struct vty_app_info pcu_vty_info;
-void *tall_pcu_ctx;
+void *tall_pcu_ctx = NULL;
 extern void *bv_tall_ctx;
 static int quit = 0;
 static int rt_prio = -1;
@@ -102,7 +115,10 @@ static void handle_options(int argc, char **argv)
 			spoof_mcc = atoi(optarg);
 			break;
 		case 'n':
-			spoof_mnc = atoi(optarg);
+			if (osmo_mnc_from_str(optarg, &spoof_mnc, &spoof_mnc_3_digits)) {
+				fprintf(stderr, "Error decoding MNC '%s'\n", optarg);
+				exit(1);
+			}
 			break;
 		case 'V':
 			print_version(1);
@@ -127,7 +143,7 @@ static void handle_options(int argc, char **argv)
 
 void sighandler(int sigset)
 {
-	if (sigset == SIGHUP || sigset == SIGPIPE)
+	if (sigset == SIGPIPE)
 		return;
 
 	fprintf(stderr, "Signal %d received.\n", sigset);
@@ -139,7 +155,6 @@ void sighandler(int sigset)
 		 * is terminated without finishing shutdown process.
 		 */
 		signal(SIGINT, SIG_DFL);
-		signal(SIGHUP, SIG_DFL);
 		signal(SIGTERM, SIG_DFL);
 		signal(SIGPIPE, SIG_DFL);
 		signal(SIGABRT, SIG_DFL);
@@ -165,9 +180,13 @@ int main(int argc, char *argv[])
 	struct gprs_rlcmac_bts *bts;
 	int rc;
 
-	tall_pcu_ctx = talloc_named_const(NULL, 1, "Osmo-PCU context");
-	if (!tall_pcu_ctx)
-		return -ENOMEM;
+	/* tall_pcu_ctx may already have been initialized in bts.cpp during early_init(). */
+	if (!tall_pcu_ctx) {
+		tall_pcu_ctx = talloc_named_const(NULL, 1, "Osmo-PCU context");
+		if (!tall_pcu_ctx)
+			return -ENOMEM;
+		osmo_init_logging2(tall_pcu_ctx, &gprs_log_info);
+	}
 
 	bts = bts_main_data();
 	bts->fc_interval = 1;
@@ -243,7 +262,6 @@ int main(int argc, char *argv[])
 
 	msgb_talloc_ctx_init(tall_pcu_ctx, 0);
 
-	osmo_init_logging(&gprs_log_info);
 	osmo_stats_init(tall_pcu_ctx);
 	rate_ctr_init(tall_pcu_ctx);
 	gprs_ns_set_log_ss(DNS);
@@ -292,12 +310,12 @@ int main(int argc, char *argv[])
 		return rc;
 
 	signal(SIGINT, sighandler);
-	signal(SIGHUP, sighandler);
 	signal(SIGTERM, sighandler);
 	signal(SIGPIPE, sighandler);
 	signal(SIGABRT, sighandler);
 	signal(SIGUSR1, sighandler);
 	signal(SIGUSR2, sighandler);
+	osmo_init_ignore_signals();
 
 	/* enable realtime priority for us */
 	if (rt_prio != -1) {
