@@ -12,9 +12,9 @@
 #include <osmocom/pcu/pcuif_proto.h>
 #include "pcu_vty.h"
 #include "gprs_rlcmac.h"
+#include <pdch.h>
 #include "bts.h"
 #include "tbf.h"
-
 #include "pcu_vty_functions.h"
 
 extern void *tall_pcu_ctx;
@@ -265,7 +265,7 @@ static int config_write_pcu(struct vty *vty)
 		}
 	}
 
-	return pcu_vty_config_write_pcu_ext(vty);
+	return CMD_SUCCESS;
 }
 
 /* per-BTS configuration */
@@ -1138,7 +1138,6 @@ int pcu_vty_init(const struct log_info *cat)
 
 	install_node(&pcu_node, config_write_pcu);
 	install_element(CONFIG_NODE, &cfg_pcu_cmd);
-	vty_install_default(PCU_NODE);
 	install_element(PCU_NODE, &cfg_pcu_egprs_cmd);
 	install_element(PCU_NODE, &cfg_pcu_no_egprs_cmd);
 	install_element(PCU_NODE, &cfg_pcu_no_two_phase_cmd);
