@@ -28,12 +28,21 @@
 #include <gprs_bssgp_pcu.h>
 #include <decoding.h>
 #include <pcu_l1_if.h>
-
+#include <gprs_coding_scheme.h>
+#include <gprs_ms.h>
+#include <llc.h>
 #include "pcu_utils.h"
 
 extern "C" {
 #include <osmocom/core/msgb.h>
 #include <osmocom/core/talloc.h>
+	#include <osmocom/core/bitvec.h>
+	#include <osmocom/core/logging.h>
+	#include <osmocom/core/rate_ctr.h>
+	#include <osmocom/core/utils.h>
+	#include <osmocom/gprs/gprs_bssgp_bss.h>
+	#include <osmocom/gprs/protocol/gsm_08_18.h>
+	#include <osmocom/gsm/tlv.h>
 }
 
 #include <errno.h>
@@ -59,7 +68,7 @@ int gprs_rlcmac_ul_tbf::assemble_forward_llc(const gprs_rlc_data *_data)
 	int i, num_frames = 0;
 	uint32_t dummy_tlli;
 
-	LOGP(DRLCMACUL, LOGL_DEBUG, "- Assembling frames: (len=%d)\n", len);
+	LOGPTBFUL(this, LOGL_DEBUG, "Assembling frames: (len=%d)\n", len);
 
 	num_frames = Decoding::rlc_data_from_ul_data(
 		rdbi, cs, data, &(frames[0]), ARRAY_SIZE(frames),
@@ -72,7 +81,7 @@ int gprs_rlcmac_ul_tbf::assemble_forward_llc(const gprs_rlc_data *_data)
 		if (frame->length) {
 			bts->rlc_ul_payload_bytes(frame->length);
 
-			LOGP(DRLCMACUL, LOGL_DEBUG, "-- Frame %d "
+			LOGPTBFUL(this, LOGL_DEBUG, "Frame %d "
 				"starts at offset %d, "
 				"length=%d, is_complete=%d\n",
 				i + 1, frame->offset, frame->length,
@@ -84,8 +93,7 @@ int gprs_rlcmac_ul_tbf::assemble_forward_llc(const gprs_rlc_data *_data)
 
 		if (frame->is_complete) {
 			/* send frame to SGSN */
-			LOGP(DRLCMACUL, LOGL_INFO, "%s complete UL frame len=%d\n",
-				tbf_name(this) , m_llc.frame_length());
+			LOGPTBFUL(this, LOGL_DEBUG, "complete UL frame len=%d\n", m_llc.frame_length());
 			snd_ul_ud();
 			bts->llc_ul_bytes(m_llc.frame_length());
 			m_llc.reset();
@@ -97,10 +105,8 @@ int gprs_rlcmac_ul_tbf::assemble_forward_llc(const gprs_rlc_data *_data)
 
 bool gprs_rlcmac_ul_tbf::ctrl_ack_to_toggle()
 {
-	if ((state_flags & (1 << GPRS_RLCMAC_FLAG_TO_UL_ACK))) {
-		state_flags &= ~(1 << GPRS_RLCMAC_FLAG_TO_UL_ACK);
+	if (check_n_clear(GPRS_RLCMAC_FLAG_TO_UL_ACK))
 		return true; /* GPRS_RLCMAC_FLAG_TO_UL_ACK was set, now cleared */
-	}
 
 	state_flags |= (1 << GPRS_RLCMAC_FLAG_TO_UL_ACK);
 	return false; /* GPRS_RLCMAC_FLAG_TO_UL_ACK was unset, now set */
@@ -109,8 +115,8 @@ bool gprs_rlcmac_ul_tbf::ctrl_ack_to_toggle()
 bool gprs_rlcmac_ul_tbf::handle_ctrl_ack()
 {
 	/* check if this control ack belongs to packet uplink ack */
-	if (ul_ack_state == GPRS_RLCMAC_UL_ACK_WAIT_ACK) {
-		ul_ack_state = GPRS_RLCMAC_UL_ACK_NONE;
+	if (ul_ack_state_is(GPRS_RLCMAC_UL_ACK_WAIT_ACK)) {
+		TBF_SET_ACK_STATE(this, GPRS_RLCMAC_UL_ACK_NONE);
 		return true;
 	}
 
@@ -126,11 +132,9 @@ struct msgb *gprs_rlcmac_ul_tbf::create_ul_ack(uint32_t fn, uint8_t ts)
 	uint32_t new_poll_fn = 0;
 
 	if (final) {
-		if (poll_state == GPRS_RLCMAC_POLL_SCHED &&
-			ul_ack_state == GPRS_RLCMAC_UL_ACK_WAIT_ACK) {
-			LOGP(DRLCMACUL, LOGL_DEBUG, "Polling is already "
-				"scheduled for %s, so we must wait for "
-				"the final uplink ack...\n", tbf_name(this));
+		if (poll_scheduled() && ul_ack_state_is(GPRS_RLCMAC_UL_ACK_WAIT_ACK)) {
+			LOGPTBFUL(this, LOGL_DEBUG,
+				  "Polling is already scheduled, so we must wait for the final uplink ack...\n");
 			return NULL;
 		}
 
@@ -162,7 +166,7 @@ struct msgb *gprs_rlcmac_ul_tbf::create_ul_ack(uint32_t fn, uint8_t ts)
 		/* waiting for final acknowledge */
 		m_final_ack_sent = 1;
 	} else
-		ul_ack_state = GPRS_RLCMAC_UL_ACK_NONE;
+		TBF_SET_ACK_STATE(this, GPRS_RLCMAC_UL_ACK_NONE);
 
 	return msg;
 }
@@ -178,7 +182,7 @@ int gprs_rlcmac_ul_tbf::rcv_data_block_acknowledged(
 
 	this->state_flags |= (1 << GPRS_RLCMAC_FLAG_UL_DATA);
 
-	LOGP(DRLCMACUL, LOGL_DEBUG, "UL DATA TFI=%d received (V(Q)=%d .. "
+	LOGPTBFUL(this, LOGL_DEBUG, "UL DATA TFI=%d received (V(Q)=%d .. "
 		"V(R)=%d)\n", rlc->tfi, this->m_window.v_q(),
 		this->m_window.v_r());
 
@@ -193,7 +197,7 @@ int gprs_rlcmac_ul_tbf::rcv_data_block_acknowledged(
 	unsigned int block_idx;
 
 	/* restart T3169 */
-	tbf_timer_start(this, 3169, bts_data()->t3169, 0);
+	T_START(this, T3169, bts_data()->t3169, 0, "acked (data)", true);
 
 	/* Increment RX-counter */
 	this->m_rx_counter++;
@@ -207,34 +211,29 @@ int gprs_rlcmac_ul_tbf::rcv_data_block_acknowledged(
 		bool need_rlc_data = false;
 		struct gprs_rlc_data *block;
 
-		LOGP(DRLCMACUL, LOGL_DEBUG,
-			"%s: Got %s RLC data block: "
-			"CV=%d, BSN=%d, SPB=%d, "
-			"PI=%d, E=%d, TI=%d, bitoffs=%d\n",
-			name(), rlc->cs.name(),
-			rdbi->cv, rdbi->bsn, rdbi->spb,
-			rdbi->pi, rdbi->e, rdbi->ti,
-			rlc->data_offs_bits[block_idx]);
+		LOGPTBFUL(this, LOGL_DEBUG,
+			  "Got %s RLC data block: CV=%d, BSN=%d, SPB=%d, PI=%d, E=%d, TI=%d, bitoffs=%d\n",
+			  rlc->cs.name(),
+			  rdbi->cv, rdbi->bsn, rdbi->spb,
+			  rdbi->pi, rdbi->e, rdbi->ti,
+			  rlc->data_offs_bits[block_idx]);
 
 		/* Check whether the block needs to be decoded */
 
 		if (!m_window.is_in_window(rdbi->bsn)) {
-			LOGP(DRLCMACUL, LOGL_DEBUG, "- BSN %d out of window "
-				"%d..%d (it's normal)\n", rdbi->bsn,
-				m_window.v_q(),
-				m_window.mod_sns(m_window.v_q() + ws - 1));
+			LOGPTBFUL(this, LOGL_DEBUG, "BSN %d out of window %d..%d (it's normal)\n",
+				  rdbi->bsn,
+				  m_window.v_q(), m_window.mod_sns(m_window.v_q() + ws - 1));
 		} else if (m_window.is_received(rdbi->bsn)) {
-			LOGP(DRLCMACUL, LOGL_DEBUG,
-				"- BSN %d already received\n", rdbi->bsn);
+			LOGPTBFUL(this, LOGL_DEBUG,
+				  "BSN %d already received\n", rdbi->bsn);
 		} else {
 			need_rlc_data = true;
 		}
 
 		if (!is_tlli_valid()) {
 			if (!rdbi->ti) {
-				LOGP(DRLCMACUL, LOGL_NOTICE,
-					"%s: Missing TLLI within UL DATA.\n",
-					name());
+				LOGPTBFUL(this, LOGL_NOTICE, "Missing TLLI within UL DATA.\n");
 				continue;
 			}
 			need_rlc_data = true;
@@ -245,9 +244,9 @@ int gprs_rlcmac_ul_tbf::rcv_data_block_acknowledged(
 
 		/* Store block and meta info to BSN buffer */
 
-		LOGP(DRLCMACUL, LOGL_DEBUG, "- BSN %d storing in window (%d..%d)\n",
-			rdbi->bsn, m_window.v_q(),
-			m_window.mod_sns(m_window.v_q() + ws - 1));
+		LOGPTBFUL(this, LOGL_DEBUG, "BSN %d storing in window (%d..%d)\n",
+			  rdbi->bsn, m_window.v_q(),
+			  m_window.mod_sns(m_window.v_q() + ws - 1));
 		block = m_rlc.block(rdbi->bsn);
 		OSMO_ASSERT(rdbi->data_len <= sizeof(block->block));
 		rlc_data = &(block->block[0]);
@@ -268,9 +267,9 @@ int gprs_rlcmac_ul_tbf::rcv_data_block_acknowledged(
 				block_idx, data, rlc_data);
 		}
 
-		LOGP(DRLCMACUL, LOGL_DEBUG,
-			"%s: data_length=%d, data=%s\n",
-			name(), block->len, osmo_hexdump(rlc_data, block->len));
+		LOGPTBFUL(this, LOGL_DEBUG,
+			  "data_length=%d, data=%s\n",
+			  block->len, osmo_hexdump(rlc_data, block->len));
 		/* Get/Handle TLLI */
 		if (rdbi->ti) {
 			num_chunks = Decoding::rlc_data_from_ul_data(
@@ -278,28 +277,27 @@ int gprs_rlcmac_ul_tbf::rcv_data_block_acknowledged(
 
 			if (num_chunks < 0) {
 				bts->decode_error();
-				LOGP(DRLCMACUL, LOGL_NOTICE,
-					"Failed to decode TLLI of %s UL DATA "
-					"TFI=%d.\n", rlc->cs.name(), rlc->tfi);
+				LOGPTBFUL(this, LOGL_NOTICE,
+					  "Failed to decode TLLI of %s UL DATA TFI=%d.\n",
+					  rlc->cs.name(), rlc->tfi);
 				m_window.invalidate_bsn(rdbi->bsn);
 				continue;
 			}
 			if (!this->is_tlli_valid()) {
 				if (!new_tlli) {
-					LOGP(DRLCMACUL, LOGL_NOTICE,
-						"%s: TLLI = 0 within UL DATA.\n",
-						name());
+					LOGPTBFUL(this, LOGL_NOTICE,
+						  "TLLI = 0 within UL DATA.\n");
 					m_window.invalidate_bsn(rdbi->bsn);
 					continue;
 				}
-				LOGP(DRLCMACUL, LOGL_INFO,
-					"Decoded premier TLLI=0x%08x of "
-					"UL DATA TFI=%d.\n", tlli(), rlc->tfi);
+				LOGPTBFUL(this, LOGL_INFO,
+					  "Decoded premier TLLI=0x%08x of UL DATA TFI=%d.\n",
+					  tlli(), rlc->tfi);
 				set_tlli_from_ul(new_tlli);
 			} else if (new_tlli && new_tlli != tlli()) {
-				LOGP(DRLCMACUL, LOGL_NOTICE, "TLLI mismatch on UL "
-					"DATA TFI=%d. (Ignoring due to contention "
-					"resolution)\n", rlc->tfi);
+				LOGPTBFUL(this, LOGL_NOTICE,
+					  "TLLI mismatch on UL DATA TFI=%d. (Ignoring due to contention resolution)\n",
+					  rlc->tfi);
 				m_window.invalidate_bsn(rdbi->bsn);
 				continue;
 			}
@@ -328,15 +326,14 @@ int gprs_rlcmac_ul_tbf::rcv_data_block_acknowledged(
 		const struct gprs_rlc_data_block_info *rdbi =
 			&block->block_info;
 
-		LOGP(DRLCMACUL, LOGL_DEBUG, "- No gaps in received block, "
-			"last block: BSN=%d CV=%d\n", rdbi->bsn,
-			rdbi->cv);
+		LOGPTBFUL(this, LOGL_DEBUG,
+			  "No gaps in received block, last block: BSN=%d CV=%d\n",
+			  rdbi->bsn, rdbi->cv);
 		if (rdbi->cv == 0) {
-			LOGP(DRLCMACUL, LOGL_DEBUG, "- Finished with UL "
-				"TBF\n");
-			set_state(GPRS_RLCMAC_FINISHED);
+			LOGPTBFUL(this, LOGL_DEBUG, "Finished with UL TBF\n");
+			TBF_SET_STATE(this, GPRS_RLCMAC_FINISHED);
 			/* Reset N3103 counter. */
-			this->m_n3103 = 0;
+			this->n_reset(N3103);
 		}
 	}
 
@@ -357,29 +354,29 @@ void gprs_rlcmac_ul_tbf::maybe_schedule_uplink_acknack(
 		(m_rx_counter % SEND_ACK_AFTER_FRAMES) == 0)
 	{
 		if (rlc->si) {
-			LOGP(DRLCMACUL, LOGL_NOTICE, "- Scheduling Ack/Nack, "
-				"because MS is stalled.\n");
+			LOGPTBFUL(this, LOGL_NOTICE,
+				  "Scheduling Ack/Nack, because MS is stalled.\n");
 		}
 		if (have_ti) {
-			LOGP(DRLCMACUL, LOGL_DEBUG, "- Scheduling Ack/Nack, "
-				"because TLLI is included.\n");
+			LOGPTBFUL(this, LOGL_DEBUG,
+				  "Scheduling Ack/Nack, because TLLI is included.\n");
 		}
 		if (state_is(GPRS_RLCMAC_FINISHED)) {
-			LOGP(DRLCMACUL, LOGL_DEBUG, "- Scheduling Ack/Nack, "
-				"because last block has CV==0.\n");
+			LOGPTBFUL(this, LOGL_DEBUG,
+				  "Scheduling Ack/Nack, because last block has CV==0.\n");
 		}
 		if ((m_rx_counter % SEND_ACK_AFTER_FRAMES) == 0) {
-			LOGP(DRLCMACUL, LOGL_DEBUG, "- Scheduling Ack/Nack, "
-				"because %d frames received.\n",
-				SEND_ACK_AFTER_FRAMES);
+			LOGPTBFUL(this, LOGL_DEBUG,
+				  "Scheduling Ack/Nack, because %d frames received.\n",
+				  SEND_ACK_AFTER_FRAMES);
 		}
-		if (ul_ack_state == GPRS_RLCMAC_UL_ACK_NONE) {
+		if (ul_ack_state_is(GPRS_RLCMAC_UL_ACK_NONE)) {
 			/* trigger sending at next RTS */
-			ul_ack_state = GPRS_RLCMAC_UL_ACK_SEND_ACK;
+			TBF_SET_ACK_STATE(this, GPRS_RLCMAC_UL_ACK_SEND_ACK);
 		} else {
 			/* already triggered */
-			LOGP(DRLCMACUL, LOGL_DEBUG, "-  Sending Ack/Nack is "
-				"already triggered, don't schedule!\n");
+			LOGPTBFUL(this, LOGL_DEBUG,
+				  "Sending Ack/Nack is already triggered, don't schedule!\n");
 		}
 	}
 }
@@ -423,21 +420,16 @@ egprs_rlc_ul_reseg_bsn_state gprs_rlcmac_ul_tbf::handle_egprs_ul_second_seg(
 
 	if (spb_status->block_status_ul &
 				EGPRS_RESEG_FIRST_SEG_RXD) {
-		LOGP(DRLCMACUL, LOGL_DEBUG,
-				"---%s: Second seg is received "
-				"first seg is already present "
-				"set the status to complete\n", name());
+		LOGPTBFUL(this, LOGL_DEBUG,
+			  "Second seg is received first seg is already present set the status to complete\n");
 		spb_status->block_status_ul = EGPRS_RESEG_DEFAULT;
 
 		block->len += Decoding::rlc_copy_to_aligned_buffer(rlc,
 			block_idx, data, rlc_data + block->len);
 		block->block_info.data_len += rdbi->data_len;
 	} else if (spb_status->block_status_ul == EGPRS_RESEG_DEFAULT) {
-		LOGP(DRLCMACUL, LOGL_DEBUG,
-				"---%s: Second seg is received "
-				"first seg is not received "
-				"set the status to second seg received\n",
-				name());
+		LOGPTBFUL(this, LOGL_DEBUG,
+			  "Second seg is received first seg is not received set the status to second seg received\n");
 
 		block->len = Decoding::rlc_copy_to_aligned_buffer(rlc,
 				block_idx, data,
@@ -460,10 +452,8 @@ egprs_rlc_ul_reseg_bsn_state gprs_rlcmac_ul_tbf::handle_egprs_ul_first_seg(
 	bts->spb_uplink_first_segment();
 
 	if (spb_status->block_status_ul & EGPRS_RESEG_SECOND_SEG_RXD) {
-		LOGP(DRLCMACUL, LOGL_DEBUG,
-				"---%s: First seg is received "
-				"second seg is already present "
-				"set the status to complete\n", name());
+		LOGPTBFUL(this, LOGL_DEBUG,
+			  "First seg is received second seg is already present set the status to complete\n");
 
 		block->len += Decoding::rlc_copy_to_aligned_buffer(rlc,
 				block_idx, data, rlc_data);
@@ -471,11 +461,8 @@ egprs_rlc_ul_reseg_bsn_state gprs_rlcmac_ul_tbf::handle_egprs_ul_first_seg(
 		block->block_info.data_len = block->len;
 		spb_status->block_status_ul = EGPRS_RESEG_DEFAULT;
 	} else if (spb_status->block_status_ul == EGPRS_RESEG_DEFAULT) {
-		LOGP(DRLCMACUL, LOGL_DEBUG,
-				"---%s: First seg is received "
-				"second seg is not received "
-				"set the status to first seg "
-				"received\n", name());
+		LOGPTBFUL(this, LOGL_DEBUG,
+			  "First seg is received second seg is not received set the status to first seg received\n");
 
 		spb_status->block_status_ul = EGPRS_RESEG_FIRST_SEG_RXD;
 		block->len = Decoding::rlc_copy_to_aligned_buffer(rlc,
@@ -491,11 +478,9 @@ egprs_rlc_ul_reseg_bsn_state gprs_rlcmac_ul_tbf::handle_egprs_ul_spb(
 {
 	const gprs_rlc_data_block_info *rdbi = &rlc->block_info[block_idx];
 
-	LOGP(DRLCMACUL, LOGL_DEBUG,
-		"--%s: Got SPB(%d)  "
-		"cs(%s) data block with BSN (%d), "
-		"TFI(%d).\n", name(), rdbi->spb,  rlc->cs.name(), rdbi->bsn,
-		rlc->tfi);
+	LOGPTBFUL(this, LOGL_DEBUG,
+		  "Got SPB(%d) cs(%s) data block with BSN (%d), TFI(%d).\n",
+		  rdbi->spb,  rlc->cs.name(), rdbi->bsn, rlc->tfi);
 
 	egprs_rlc_ul_reseg_bsn_state assemble_status = EGPRS_RESEG_INVALID;
 
@@ -507,10 +492,9 @@ egprs_rlc_ul_reseg_bsn_state gprs_rlcmac_ul_tbf::handle_egprs_ul_spb(
 		assemble_status = handle_egprs_ul_second_seg(rlc,
 						block, data, block_idx);
 	else {
-		LOGP(DRLCMACUL, LOGL_ERROR,
-			"--%s: spb(%d) Not supported SPB for this EGPRS "
-			"configuration\n",
-			name(), rdbi->spb);
+		LOGPTBFUL(this, LOGL_ERROR,
+			  "spb(%d) Not supported SPB for this EGPRS configuration\n",
+			  rdbi->spb);
 	}
 
 	/*
@@ -521,23 +505,20 @@ egprs_rlc_ul_reseg_bsn_state gprs_rlcmac_ul_tbf::handle_egprs_ul_spb(
 		switch (GprsCodingScheme::Scheme(rlc->cs)) {
 		case GprsCodingScheme::MCS3 :
 			block->cs_last = GprsCodingScheme::MCS6;
-			LOGP(DRLCMACUL, LOGL_DEBUG,
-				"--%s: Upgrading to MCS6\n", name());
+			LOGPTBFUL(this, LOGL_DEBUG, "Upgrading to MCS6\n");
 			break;
 		case GprsCodingScheme::MCS2 :
 			block->cs_last = GprsCodingScheme::MCS5;
-			LOGP(DRLCMACUL, LOGL_DEBUG,
-				"--%s: Upgrading to MCS5\n", name());
+			LOGPTBFUL(this, LOGL_DEBUG, "Upgrading to MCS5\n");
 			break;
 		case GprsCodingScheme::MCS1 :
-			LOGP(DRLCMACUL, LOGL_DEBUG,
-				"--%s: Upgrading to MCS4\n", name());
+			LOGPTBFUL(this, LOGL_DEBUG, "Upgrading to MCS4\n");
 			block->cs_last = GprsCodingScheme::MCS4;
 			break;
 		default:
-			LOGP(DRLCMACUL, LOGL_ERROR,
-				"--%s: cs(%s) Error in Upgrading to higher MCS\n",
-				name(), rlc->cs.name());
+			LOGPTBFUL(this, LOGL_ERROR,
+				  "cs(%s) Error in Upgrading to higher MCS\n",
+				  rlc->cs.name());
 			break;
 		}
 	}
@@ -610,21 +591,10 @@ void gprs_rlcmac_ul_tbf::update_coding_scheme_counter_ul(const GprsCodingScheme
 	}
 }
 
-void gprs_rlcmac_ul_tbf::egprs_calc_ulwindow_size()
+void gprs_rlcmac_ul_tbf::set_window_size()
 {
-	struct gprs_rlcmac_bts *bts_data = bts->bts_data();
-	unsigned int num_pdch = pcu_bitcount(ul_slots());
-	unsigned int ws = bts_data->ws_base + num_pdch * bts_data->ws_pdch;
-	ws = (ws / 32) * 32;
-	ws = OSMO_MAX(64, ws);
-
-	if (num_pdch == 1)
-		ws = OSMO_MIN(192, ws);
-	else
-		ws = OSMO_MIN(128 * num_pdch, ws);
-
-	LOGP(DRLCMAC, LOGL_INFO, "%s: Setting EGPRS window size to %d, base(%d) slots(%d) ws_pdch(%d)\n",
-		name(), ws, bts_data->ws_base, num_pdch, bts_data->ws_pdch);
-
+	uint16_t ws = egprs_window_size(bts->bts_data(), ul_slots());
+	LOGPTBFUL(this, LOGL_INFO, "setting EGPRS UL window size to %u, base(%u) slots(%u) ws_pdch(%u)\n",
+		  ws, bts->bts_data()->ws_base, pcu_bitcount(ul_slots()), bts->bts_data()->ws_pdch);
 	m_window.set_ws(ws);
 }
