@@ -26,6 +26,7 @@
 #include <stdio.h>
 
 extern "C" {
+#include "mslot_class.h"
 #include <osmocom/core/application.h>
 #include <osmocom/core/msgb.h>
 #include <osmocom/core/talloc.h>
@@ -35,11 +36,12 @@ extern "C" {
 /* globals used by the code */
 void *tall_pcu_ctx;
 int16_t spoof_mnc = 0, spoof_mcc = 0;
+bool spoof_mnc_3_digits = false;
 
 static gprs_rlcmac_tbf *tbf_alloc(struct gprs_rlcmac_bts *bts,
 		GprsMs *ms, gprs_rlcmac_tbf_direction dir,
 		uint8_t use_trx,
-		uint8_t ms_class, uint8_t egprs_ms_class, uint8_t single_slot)
+		uint8_t ms_class, uint8_t egprs_ms_class, bool single_slot)
 {
 	if (dir == GPRS_RLCMAC_UL_TBF)
 		return tbf_alloc_ul_tbf(bts, ms, use_trx,
@@ -166,8 +168,11 @@ static void test_alloc_a()
 	test_alloc_a(GPRS_RLCMAC_UL_TBF, 0x1e, 28);
 }
 
-static void dump_assignment(struct gprs_rlcmac_tbf *tbf, const char *dir)
+static void dump_assignment(struct gprs_rlcmac_tbf *tbf, const char *dir, bool verbose)
 {
+	if (!verbose)
+		return;
+
 	for (size_t i = 0; i < ARRAY_SIZE(tbf->pdch); ++i)
 		if (tbf->pdch[i])
 			printf("PDCH[%zu] is used for %s\n", i, dir);
@@ -175,8 +180,165 @@ static void dump_assignment(struct gprs_rlcmac_tbf *tbf, const char *dir)
 	printf("PDCH[%d] is first common for %s\n", tbf->first_common_ts, dir);
 }
 
-static void test_alloc_b(int ms_class)
+#define ENABLE_PDCH(ts_no, enable_flag, trx)	\
+		if (enable_flag)		\
+			trx->pdch[ts_no].enable();
+
+static inline void enable_ts_on_bts(struct gprs_rlcmac_bts *bts,
+				    bool ts0, bool ts1, bool ts2, bool ts3, bool ts4, bool ts5, bool ts6, bool ts7)
 {
+	struct gprs_rlcmac_trx *trx = &bts->trx[0];
+
+	ENABLE_PDCH(0, ts0, trx);
+	ENABLE_PDCH(1, ts1, trx);
+	ENABLE_PDCH(2, ts2, trx);
+	ENABLE_PDCH(3, ts3, trx);
+	ENABLE_PDCH(4, ts4, trx);
+	ENABLE_PDCH(5, ts5, trx);
+	ENABLE_PDCH(6, ts6, trx);
+	ENABLE_PDCH(7, ts7, trx);
+}
+
+static inline bool test_alloc_b_ul_dl(bool ts0, bool ts1, bool ts2, bool ts3, bool ts4, bool ts5, bool ts6, bool ts7,
+				      uint8_t ms_class, bool verbose)
+{
+	BTS the_bts;
+	struct gprs_rlcmac_bts *bts = the_bts.bts_data();
+	gprs_rlcmac_ul_tbf *ul_tbf;
+	gprs_rlcmac_dl_tbf *dl_tbf;
+
+	if (verbose)
+		printf("Testing UL then DL assignment.\n");
+
+	bts->alloc_algorithm = alloc_algorithm_b;
+
+	enable_ts_on_bts(bts, ts0, ts1, ts2, ts3, ts4, ts5, ts6, ts7);
+
+	ul_tbf = tbf_alloc_ul_tbf(bts, NULL, -1, ms_class, 0, true);
+	if (!ul_tbf)
+		return false;
+
+	OSMO_ASSERT(ul_tbf->ms());
+	OSMO_ASSERT(ul_tbf->ms()->current_trx());
+
+	dump_assignment(ul_tbf, "UL", verbose);
+
+	/* assume final ack has not been sent */
+	dl_tbf = tbf_alloc_dl_tbf(bts, ul_tbf->ms(), ul_tbf->ms()->current_trx()->trx_no, ms_class, 0,
+				  false);
+	if (!dl_tbf)
+		return false;
+
+	dump_assignment(dl_tbf, "DL", verbose);
+
+	OSMO_ASSERT(dl_tbf->first_common_ts == ul_tbf->first_common_ts);
+
+	check_tfi_usage(&the_bts);
+
+	tbf_free(dl_tbf);
+	tbf_free(ul_tbf);
+
+	return true;
+}
+
+static inline bool test_alloc_b_dl_ul(bool ts0, bool ts1, bool ts2, bool ts3, bool ts4, bool ts5, bool ts6, bool ts7,
+				      uint8_t ms_class, bool verbose)
+{
+	BTS the_bts;
+	struct gprs_rlcmac_bts *bts = the_bts.bts_data();
+	gprs_rlcmac_ul_tbf *ul_tbf;
+	gprs_rlcmac_dl_tbf *dl_tbf;
+
+	if (verbose)
+		printf("Testing DL then UL assignment followed by update\n");
+
+	bts->alloc_algorithm = alloc_algorithm_b;
+
+	enable_ts_on_bts(bts, ts0, ts1, ts2, ts3, ts4, ts5, ts6, ts7);
+
+	dl_tbf = tbf_alloc_dl_tbf(bts, NULL, -1, ms_class, 0, true);
+	if (!dl_tbf)
+		return false;
+
+	dl_tbf->update_ms(0x23, GPRS_RLCMAC_DL_TBF);
+
+	OSMO_ASSERT(dl_tbf->ms());
+	OSMO_ASSERT(dl_tbf->ms()->current_trx());
+
+	dump_assignment(dl_tbf, "DL", verbose);
+
+	ul_tbf = tbf_alloc_ul_tbf(bts, dl_tbf->ms(), dl_tbf->ms()->current_trx()->trx_no, ms_class, 0,
+				  false);
+	if (!ul_tbf)
+		return false;
+
+	ul_tbf->update_ms(0x23, GPRS_RLCMAC_UL_TBF);
+	ul_tbf->m_contention_resolution_done = 1;
+
+	dump_assignment(ul_tbf, "UL", verbose);
+
+	OSMO_ASSERT(dl_tbf->first_common_ts == ul_tbf->first_common_ts);
+
+	/* now update the dl_tbf */
+	dl_tbf->update();
+	dump_assignment(dl_tbf, "DL", verbose);
+	OSMO_ASSERT(dl_tbf->first_common_ts == ul_tbf->first_common_ts);
+
+	check_tfi_usage(&the_bts);
+
+	tbf_free(dl_tbf);
+	tbf_free(ul_tbf);
+
+	return true;
+}
+
+static inline bool test_alloc_b_jolly(uint8_t ms_class)
+{
+	BTS the_bts;
+	struct gprs_rlcmac_bts *bts = the_bts.bts_data();
+	int tfi;
+	uint8_t trx_no;
+	gprs_rlcmac_tbf *ul_tbf, *dl_tbf;
+
+	printf("Testing jolly example\n");
+
+	bts->alloc_algorithm = alloc_algorithm_b;
+
+	enable_ts_on_bts(bts, false, true, true, true, true, false, false, false);
+
+	tfi = the_bts.tfi_find_free(GPRS_RLCMAC_UL_TBF, &trx_no, -1);
+	OSMO_ASSERT(tfi >= 0);
+	ul_tbf = tbf_alloc_ul_tbf(bts, NULL, .1, ms_class, 0, false);
+	if (!ul_tbf)
+		return false;
+
+	OSMO_ASSERT(ul_tbf->ms());
+	OSMO_ASSERT(ul_tbf->ms()->current_trx());
+	trx_no = ul_tbf->ms()->current_trx()->trx_no;
+	dump_assignment(ul_tbf, "UL", true);
+
+	/* assume final ack has not been sent */
+	dl_tbf = tbf_alloc_dl_tbf(bts, ul_tbf->ms(), trx_no, ms_class, 0,
+				  false);
+	if (!dl_tbf)
+		return false;
+
+	dump_assignment(dl_tbf, "DL", true);
+
+	OSMO_ASSERT(dl_tbf->first_common_ts == ul_tbf->first_common_ts);
+
+	check_tfi_usage(&the_bts);
+
+	tbf_free(dl_tbf);
+	tbf_free(ul_tbf);
+
+	return true;
+}
+
+static void test_alloc_b_for_ms(uint8_t ms_class)
+{
+	bool rc;
+
 	printf("Going to test multislot assignment MS_CLASS=%d\n", ms_class);
 	/*
 	 * PDCH is on TS 6,7,8 and we start with a UL allocation and
@@ -184,143 +346,26 @@ static void test_alloc_b(int ms_class)
 	 *
 	 * Uplink assigned and still available..
 	 */
-	{
-		BTS the_bts;
-		struct gprs_rlcmac_bts *bts;
-		struct gprs_rlcmac_trx *trx;
-		uint8_t trx_no;
 
-		gprs_rlcmac_tbf *ul_tbf, *dl_tbf;
-
-		printf("Testing UL then DL assignment.\n");
-
-		bts = the_bts.bts_data();
-		bts->alloc_algorithm = alloc_algorithm_b;
-
-		trx = &bts->trx[0];
-		trx->pdch[5].enable();
-		trx->pdch[6].enable();
-		trx->pdch[7].enable();
-
-		ul_tbf = tbf_alloc_ul_tbf(bts, NULL, -1, ms_class, 0, 1);
-		OSMO_ASSERT(ul_tbf);
-		OSMO_ASSERT(ul_tbf->ms());
-		OSMO_ASSERT(ul_tbf->ms()->current_trx());
-		trx_no = ul_tbf->ms()->current_trx()->trx_no;
-		dump_assignment(ul_tbf, "UL");
-
-		/* assume final ack has not been sent */
-		dl_tbf = tbf_alloc_dl_tbf(bts, ul_tbf->ms(), trx_no, ms_class, 0, 0);
-		OSMO_ASSERT(dl_tbf);
-		dump_assignment(dl_tbf, "DL");
-
-		OSMO_ASSERT(dl_tbf->first_common_ts == ul_tbf->first_common_ts);
-
-		check_tfi_usage(&the_bts);
-
-		tbf_free(dl_tbf);
-		tbf_free(ul_tbf);
-	}
+	rc = test_alloc_b_ul_dl(false, false, false, false, false, true, true, true, ms_class, true);
+	if (!rc)
+		return;
 
 	/**
 	 * Test with the other order.. first DL and then UL
 	 */
-	{
-		BTS the_bts;
-		struct gprs_rlcmac_bts *bts;
-		struct gprs_rlcmac_trx *trx;
-		uint8_t trx_no;
-
-		gprs_rlcmac_ul_tbf *ul_tbf;
-		gprs_rlcmac_dl_tbf *dl_tbf;
-
-		printf("Testing DL then UL assignment followed by update\n");
-
-		bts = the_bts.bts_data();
-		bts->alloc_algorithm = alloc_algorithm_b;
-
-		trx = &bts->trx[0];
-		trx->pdch[5].enable();
-		trx->pdch[6].enable();
-		trx->pdch[7].enable();
-
-		dl_tbf = tbf_alloc_dl_tbf(bts, NULL, -1, ms_class, 0, 1);
-		dl_tbf->update_ms(0x23, GPRS_RLCMAC_DL_TBF);
-		OSMO_ASSERT(dl_tbf);
-		OSMO_ASSERT(dl_tbf->ms());
-		OSMO_ASSERT(dl_tbf->ms()->current_trx());
-		trx_no = dl_tbf->ms()->current_trx()->trx_no;
-		dump_assignment(dl_tbf, "DL");
-
-		ul_tbf = tbf_alloc_ul_tbf(bts, dl_tbf->ms(), trx_no, ms_class, 0, 0);
-		ul_tbf->update_ms(0x23, GPRS_RLCMAC_UL_TBF);
-		ul_tbf->m_contention_resolution_done = 1;
-		OSMO_ASSERT(ul_tbf);
-		dump_assignment(ul_tbf, "UL");
-
-		OSMO_ASSERT(dl_tbf->first_common_ts == ul_tbf->first_common_ts);
-
-		/* now update the dl_tbf */
-		dl_tbf->update();
-		dump_assignment(dl_tbf, "DL");
-		OSMO_ASSERT(dl_tbf->first_common_ts == ul_tbf->first_common_ts);
-
-		check_tfi_usage(&the_bts);
-
-		tbf_free(dl_tbf);
-		tbf_free(ul_tbf);
-	}
+	rc = test_alloc_b_dl_ul(false, false, false, false, false, true, true, true, ms_class, true);
+	if (!rc)
+		return;
 
 	/* Andreas osmocom-pcu example */
-	{
-		BTS the_bts;
-		struct gprs_rlcmac_bts *bts;
-		struct gprs_rlcmac_trx *trx;
-		int tfi;
-		uint8_t trx_no;
-
-		gprs_rlcmac_tbf *ul_tbf, *dl_tbf;
-
-		printf("Testing jolly example\n");
-
-		bts = the_bts.bts_data();
-		bts->alloc_algorithm = alloc_algorithm_b;
-
-		trx = &bts->trx[0];
-		trx->pdch[1].enable();
-		trx->pdch[2].enable();
-		trx->pdch[3].enable();
-		trx->pdch[4].enable();
-
-		tfi = the_bts.tfi_find_free(GPRS_RLCMAC_UL_TBF, &trx_no, -1);
-		OSMO_ASSERT(tfi >= 0);
-		ul_tbf = tbf_alloc_ul_tbf(bts, NULL, .1, ms_class, 0, 0);
-		OSMO_ASSERT(ul_tbf);
-		OSMO_ASSERT(ul_tbf->ms());
-		OSMO_ASSERT(ul_tbf->ms()->current_trx());
-		trx_no = ul_tbf->ms()->current_trx()->trx_no;
-		dump_assignment(ul_tbf, "UL");
-
-		/* assume final ack has not been sent */
-		dl_tbf = tbf_alloc_dl_tbf(bts, ul_tbf->ms(), trx_no, ms_class, 0, 0);
-		OSMO_ASSERT(dl_tbf);
-		dump_assignment(dl_tbf, "DL");
-
-		OSMO_ASSERT(dl_tbf->first_common_ts == ul_tbf->first_common_ts);
-
-		check_tfi_usage(&the_bts);
-
-		tbf_free(dl_tbf);
-		tbf_free(ul_tbf);
-	}
+	test_alloc_b_jolly(ms_class);
 }
 
-#define ENABLE_PDCH(ts_no, enable_flag, trx)	\
-		if (enable_flag)		\
-			trx->pdch[ts_no].enable();
-
-static void test_alloc_b(bool ts0, bool ts1, bool ts2, bool ts3, bool ts4, bool ts5, bool ts6, bool ts7, int ms_class)
+static void test_alloc_mass(bool ts0, bool ts1, bool ts2, bool ts3, bool ts4, bool ts5, bool ts6, bool ts7, int ms_class)
 {
+	bool rc;
+
 	/* we can test the allocation failures differently */
 	if (!ts0 && !ts1 && !ts2 && !ts3 && !ts4 && !ts5 && !ts6 && !ts7)
 		return;
@@ -336,97 +381,14 @@ static void test_alloc_b(bool ts0, bool ts1, bool ts2, bool ts3, bool ts4, bool
 		ts7 ? 'O' : 'x', ms_class);
 	fflush(stdout);
 
-	{
-		BTS the_bts;
-		struct gprs_rlcmac_bts *bts;
-		struct gprs_rlcmac_trx *trx;
-		uint8_t trx_no;
-
-		gprs_rlcmac_tbf *ul_tbf, *dl_tbf;
-
-		bts = the_bts.bts_data();
-		bts->alloc_algorithm = alloc_algorithm_b;
-
-		trx = &bts->trx[0];
-		ENABLE_PDCH(0, ts0, trx);
-		ENABLE_PDCH(1, ts1, trx);
-		ENABLE_PDCH(2, ts2, trx);
-		ENABLE_PDCH(3, ts3, trx);
-		ENABLE_PDCH(4, ts4, trx);
-		ENABLE_PDCH(5, ts5, trx);
-		ENABLE_PDCH(6, ts6, trx);
-		ENABLE_PDCH(7, ts7, trx);
-
-		ul_tbf = tbf_alloc_ul_tbf(bts, NULL, -1, ms_class, 0, 1);
-		OSMO_ASSERT(ul_tbf->ms());
-		OSMO_ASSERT(ul_tbf->ms()->current_trx());
-		trx_no = ul_tbf->ms()->current_trx()->trx_no;
-		OSMO_ASSERT(ul_tbf);
-
-		/* assume final ack has not been sent */
-		dl_tbf = tbf_alloc_dl_tbf(bts, ul_tbf->ms(), trx_no, ms_class, 0, 0);
-		OSMO_ASSERT(dl_tbf);
-
-		/* verify that both are on the same ts */
-		OSMO_ASSERT(dl_tbf->first_common_ts == ul_tbf->first_common_ts);
-
-		check_tfi_usage(&the_bts);
-
-		tbf_free(dl_tbf);
-		tbf_free(ul_tbf);
-	}
+	rc = test_alloc_b_ul_dl(ts0, ts1, ts2, ts3, ts4, ts5, ts6, ts7, ms_class, false);
+	if (!rc)
+		return;
 
 	/**
 	 * Test with the other order.. first DL and then UL
 	 */
-	{
-		BTS the_bts;
-		struct gprs_rlcmac_bts *bts;
-		struct gprs_rlcmac_trx *trx;
-		uint8_t trx_no;
-
-		gprs_rlcmac_ul_tbf *ul_tbf;
-		gprs_rlcmac_dl_tbf *dl_tbf;
-
-		bts = the_bts.bts_data();
-		bts->alloc_algorithm = alloc_algorithm_b;
-
-		trx = &bts->trx[0];
-		ENABLE_PDCH(0, ts0, trx);
-		ENABLE_PDCH(1, ts1, trx);
-		ENABLE_PDCH(2, ts2, trx);
-		ENABLE_PDCH(3, ts3, trx);
-		ENABLE_PDCH(4, ts4, trx);
-		ENABLE_PDCH(5, ts5, trx);
-		ENABLE_PDCH(6, ts6, trx);
-		ENABLE_PDCH(7, ts7, trx);
-
-		dl_tbf = tbf_alloc_dl_tbf(bts, NULL, -1, ms_class, 0, 1);
-		OSMO_ASSERT(dl_tbf);
-		OSMO_ASSERT(dl_tbf->ms());
-		OSMO_ASSERT(dl_tbf->ms()->current_trx());
-		trx_no = dl_tbf->ms()->current_trx()->trx_no;
-		dl_tbf->update_ms(0x23, GPRS_RLCMAC_DL_TBF);
-
-		ul_tbf = tbf_alloc_ul_tbf(bts, dl_tbf->ms(), trx_no, ms_class, 0, 0);
-		OSMO_ASSERT(ul_tbf);
-		ul_tbf->update_ms(0x23, GPRS_RLCMAC_UL_TBF);
-		ul_tbf->m_contention_resolution_done = 1;
-
-		OSMO_ASSERT(dl_tbf->first_common_ts == ul_tbf->first_common_ts);
-
-		/* now update the dl_tbf */
-		dl_tbf->update();
-		OSMO_ASSERT(dl_tbf->first_common_ts == ul_tbf->first_common_ts);
-
-		OSMO_ASSERT(ul_tbf->ms_class() == ms_class);
-		OSMO_ASSERT(dl_tbf->ms_class() == ms_class);
-
-		check_tfi_usage(&the_bts);
-
-		tbf_free(dl_tbf);
-		tbf_free(ul_tbf);
-	}
+	test_alloc_b_dl_ul(ts0, ts1, ts2, ts3, ts4, ts5, ts6, ts7, ms_class, false);
 }
 
 static void test_all_alloc_b()
@@ -440,22 +402,20 @@ static void test_all_alloc_b()
        for (uint8_t ts5 = 0; ts5 < 2; ++ts5)
         for (uint8_t ts6 = 0; ts6 < 2; ++ts6)
          for (uint8_t ts7 = 0; ts7 < 2; ++ts7)
-	  for (int ms_class = 0; ms_class < 30; ++ms_class)
-		test_alloc_b(ts0, ts1, ts2, ts3, ts4, ts5, ts6, ts7, ms_class);
+	  for (int ms_class = 0; ms_class < mslot_class_max(); ++ms_class)
+		test_alloc_mass(ts0, ts1, ts2, ts3, ts4, ts5, ts6, ts7, ms_class);
 }
 
 static void test_alloc_b()
 {
-	for (int i = 0; i < 30; ++i)
-		test_alloc_b(i);
+	for (int i = 0; i < mslot_class_max(); ++i)
+		test_alloc_b_for_ms(i);
 
 	test_all_alloc_b();
 }
 
-typedef int (*algo_t)(struct gprs_rlcmac_bts *bts,
-		struct GprsMs *ms,
-		struct gprs_rlcmac_tbf *tbf, uint32_t cust, uint8_t single,
-		int use_trx);
+typedef int (*algo_t)(struct gprs_rlcmac_bts *bts, struct GprsMs *ms, struct gprs_rlcmac_tbf *tbf, bool single,
+		      int8_t use_trx);
 
 static char get_dir_char(uint8_t mask, uint8_t tx, uint8_t rx, uint8_t busy)
 {
@@ -475,6 +435,19 @@ enum test_mode {
 	TEST_MODE_UL_AFTER_DL,
 };
 
+static inline char *test_mode_descr(enum test_mode t)
+{
+	switch (t) {
+	case TEST_MODE_UL_ONLY: return (char*)"UL only";
+	case TEST_MODE_DL_ONLY: return (char*)"DL only";
+	case TEST_MODE_UL_AND_DL: return (char*)"UL and DL";
+	case TEST_MODE_DL_AND_UL: return (char*)"DL and UL";
+	case TEST_MODE_DL_AFTER_UL: return (char*)"DL after UL";
+	case TEST_MODE_UL_AFTER_DL: return (char*)"UL after DL";
+	default: return NULL;
+	}
+}
+
 static GprsMs *alloc_tbfs(BTS *the_bts, GprsMs *ms, unsigned ms_class,
 	enum test_mode mode)
 {
@@ -497,7 +470,7 @@ static GprsMs *alloc_tbfs(BTS *the_bts, GprsMs *ms, unsigned ms_class,
 	case TEST_MODE_UL_AND_DL:
 		if (ms && ms->ul_tbf())
 			tbf_free(ms->ul_tbf());
-		tbf = tbf_alloc_ul_tbf(bts, ms, trx_no, ms_class, 0, 0);
+		tbf = tbf_alloc_ul_tbf(bts, ms, trx_no, ms_class, 0, false);
 		if (tbf == NULL)
 			return NULL;
 		break;
@@ -506,7 +479,7 @@ static GprsMs *alloc_tbfs(BTS *the_bts, GprsMs *ms, unsigned ms_class,
 	case TEST_MODE_DL_AND_UL:
 		if (ms && ms->dl_tbf())
 			tbf_free(ms->dl_tbf());
-		tbf = tbf_alloc_dl_tbf(bts, ms, trx_no, ms_class, 0, 0);
+		tbf = tbf_alloc_dl_tbf(bts, ms, trx_no, ms_class, 0, false);
 		if (tbf == NULL)
 			return NULL;
 	}
@@ -613,17 +586,17 @@ static unsigned alloc_many_tbfs(BTS *the_bts, unsigned min_class,
 				continue;
 
 			if (ul_tbf &&
-				pdch->assigned_tfi(GPRS_RLCMAC_DL_TBF) != 0xffffffff)
+				pdch->assigned_tfi(GPRS_RLCMAC_DL_TBF) != NO_FREE_TFI)
 				continue;
 
 			if (dl_tbf &&
-				pdch->assigned_tfi(GPRS_RLCMAC_UL_TBF) != 0xffffffff)
+				pdch->assigned_tfi(GPRS_RLCMAC_UL_TBF) != NO_FREE_TFI)
 				continue;
 
 			busy_slots |= 1 << i;
 		}
 
-		printf(" TBF[%d] class %d reserves %c%c%c%c%c%c%c%c\n",
+		printf(" TBF[%d] class %d reserves " OSMO_BIT_SPEC "\n",
 			tfi, ms_class,
 			get_dir_char(0x01, ul_slots, dl_slots, busy_slots),
 			get_dir_char(0x02, ul_slots, dl_slots, busy_slots),
@@ -660,7 +633,8 @@ static void test_successive_allocation(algo_t algo, unsigned min_class,
 	struct gprs_rlcmac_trx *trx;
 	unsigned counter;
 
-	printf("Going to test assignment with many TBF, %s\n", text);
+	printf("Going to test assignment with many TBF, algorithm %s class %u..%u (%s)\n",
+	       text, min_class, max_class, test_mode_descr(mode));
 
 	bts = the_bts.bts_data();
 	bts->alloc_algorithm = algo;
@@ -674,9 +648,11 @@ static void test_successive_allocation(algo_t algo, unsigned min_class,
 
 	counter = alloc_many_tbfs(&the_bts, min_class, max_class, mode);
 
-	printf("  Successfully allocated %d UL TBFs\n", counter);
+	printf("  Successfully allocated %u UL TBFs, algorithm %s class %u..%u (%s)\n",
+	       counter, text, min_class, max_class, test_mode_descr(mode));
 	if (counter != expect_num)
-		fprintf(stderr, "  Expected %d TBFs for %s\n", expect_num, text);
+		fprintf(stderr, "  Expected %u TBFs (got %u), algorithm %s class %u..%u (%s)\n",
+			expect_num, counter, text, min_class, max_class, test_mode_descr(mode));
 
 	OSMO_ASSERT(counter == expect_num);
 
@@ -698,7 +674,7 @@ static void test_many_connections(algo_t algo, unsigned expect_num,
 		TEST_MODE_DL_ONLY,
 	};
 
-	printf("Going to test assignment with many connections, %s\n", text);
+	printf("Going to test assignment with many connections, algorithm %s\n", text);
 
 	bts = the_bts.bts_data();
 	bts->alloc_algorithm = algo;
@@ -711,7 +687,7 @@ static void test_many_connections(algo_t algo, unsigned expect_num,
 	trx->pdch[7].enable();
 
 	for (i = 0; i < ARRAY_SIZE(mode_seq); i += 1) {
-		counter1 = alloc_many_tbfs(&the_bts, 1, 29, mode_seq[i]);
+		counter1 = alloc_many_tbfs(&the_bts, 1, mslot_class_max(), mode_seq[i]);
 		fprintf(stderr, "  Allocated %d TBFs (previously %d)\n",
 			counter1, counter2);
 
@@ -733,67 +709,33 @@ static void test_many_connections(algo_t algo, unsigned expect_num,
 
 	printf("  Successfully allocated %d TBFs\n", counter1);
 	if (counter1 != (int)expect_num)
-		fprintf(stderr, "  Expected %d TBFs for %s\n", expect_num, text);
+		fprintf(stderr, "  Expected %d TBFs (got %d) for algorithm %s\n", expect_num, counter1, text);
 
 	OSMO_ASSERT(expect_num == (unsigned)counter1);
 }
 
-static void test_successive_allocation()
+static inline void test_a_b_dyn(enum test_mode mode, uint8_t exp_A, uint8_t exp_B, uint8_t exp_dyn)
 {
-	test_successive_allocation(alloc_algorithm_a, 1, 1, TEST_MODE_UL_AND_DL,
-		35, "algorithm A (UL and DL)");
-	test_successive_allocation(alloc_algorithm_b, 10, 10, TEST_MODE_UL_AND_DL,
-		32, "algorithm B class 10 (UL and DL)");
-	test_successive_allocation(alloc_algorithm_b, 12, 12, TEST_MODE_UL_AND_DL,
-		32, "algorithm B class 12 (UL and DL)");
-	test_successive_allocation(alloc_algorithm_b, 1, 12, TEST_MODE_UL_AND_DL,
-		32, "algorithm B class 1-12 (UL and DL)");
-	test_successive_allocation(alloc_algorithm_b, 1, 29, TEST_MODE_UL_AND_DL,
-		32, "algorithm B class 1-29 (UL and DL)");
-	test_successive_allocation(alloc_algorithm_dynamic, 1, 29, TEST_MODE_UL_AND_DL,
-		35, "algorithm dynamic class 1-29 (UL and DL)");
-
-	test_successive_allocation(alloc_algorithm_a, 1, 1, TEST_MODE_DL_AND_UL,
-		35, "algorithm A (DL and UL)");
-	test_successive_allocation(alloc_algorithm_b, 10, 10, TEST_MODE_DL_AND_UL,
-		32, "algorithm B class 10 (DL and UL)");
-	test_successive_allocation(alloc_algorithm_dynamic, 10, 10, TEST_MODE_DL_AND_UL,
-		32, "algorithm dynamic class 10 (DL and UL)");
-
-	test_successive_allocation(alloc_algorithm_a, 1, 1, TEST_MODE_DL_AFTER_UL,
-		160, "algorithm A (DL after UL)");
-	test_successive_allocation(alloc_algorithm_b, 10, 10, TEST_MODE_DL_AFTER_UL,
-		32, "algorithm B class 10 (DL after UL)");
-	test_successive_allocation(alloc_algorithm_dynamic, 10, 10, TEST_MODE_DL_AFTER_UL,
-		95, "algorithm dynamic class 10 (DL after UL)");
-
-	test_successive_allocation(alloc_algorithm_a, 1, 1, TEST_MODE_UL_AFTER_DL,
-		35, "algorithm A (UL after DL)");
-	test_successive_allocation(alloc_algorithm_b, 10, 10, TEST_MODE_UL_AFTER_DL,
-		32, "algorithm B class 10 (UL after DL)");
-	test_successive_allocation(alloc_algorithm_dynamic, 10, 10, TEST_MODE_UL_AFTER_DL,
-		35, "algorithm dynamic class 10 (UL after DL)");
-
-	test_successive_allocation(alloc_algorithm_a, 1, 1, TEST_MODE_UL_ONLY,
-		35, "algorithm A (UL only)");
-	test_successive_allocation(alloc_algorithm_b, 10, 10, TEST_MODE_UL_ONLY,
-		32, "algorithm B class 10 (UL only)");
-	test_successive_allocation(alloc_algorithm_dynamic, 10, 10, TEST_MODE_UL_ONLY,
-		35, "algorithm dynamic class 10 (UL only)");
-
-	test_successive_allocation(alloc_algorithm_a, 1, 1, TEST_MODE_DL_ONLY,
-		160, "algorithm A (DL ONLY)");
-	test_successive_allocation(alloc_algorithm_b, 10, 10, TEST_MODE_DL_ONLY,
-		32, "algorithm B class 10 (DL ONLY)");
-	test_successive_allocation(alloc_algorithm_dynamic, 10, 10, TEST_MODE_DL_ONLY,
-		101, "algorithm dynamic class 10 (DL ONLY)");
+	test_successive_allocation(alloc_algorithm_a,        1,  1, mode, exp_A,   "A");
+	test_successive_allocation(alloc_algorithm_b,       10, 10, mode, exp_B,   "B");
+	test_successive_allocation(alloc_algorithm_dynamic, 10, 10, mode, exp_dyn, "dynamic");
 }
 
-static void test_many_connections()
+static void test_successive_allocations()
 {
-	test_many_connections(alloc_algorithm_a, 160, "algorithm A");
-	test_many_connections(alloc_algorithm_b, 32, "algorithm B");
-	test_many_connections(alloc_algorithm_dynamic, 160, "algorithm dynamic");
+	test_successive_allocation(alloc_algorithm_a,       1,  1, TEST_MODE_UL_AND_DL, 35, "A");
+	test_successive_allocation(alloc_algorithm_b,      10, 10, TEST_MODE_UL_AND_DL, 32, "B");
+	test_successive_allocation(alloc_algorithm_b,      12, 12, TEST_MODE_UL_AND_DL, 32, "B");
+
+	test_successive_allocation(alloc_algorithm_b,       1,                12, TEST_MODE_UL_AND_DL, 32, "B");
+	test_successive_allocation(alloc_algorithm_b,       1, mslot_class_max(), TEST_MODE_UL_AND_DL, 32, "B");
+	test_successive_allocation(alloc_algorithm_dynamic, 1, mslot_class_max(), TEST_MODE_UL_AND_DL, 35, "dynamic");
+
+	test_a_b_dyn(TEST_MODE_DL_AND_UL,    35, 32,  32);
+	test_a_b_dyn(TEST_MODE_DL_AFTER_UL, 160, 32,  95);
+	test_a_b_dyn(TEST_MODE_UL_AFTER_DL,  35, 32,  35);
+	test_a_b_dyn(TEST_MODE_UL_ONLY,      35, 32,  35);
+	test_a_b_dyn(TEST_MODE_DL_ONLY,     160, 32, 101);
 }
 
 static void test_2_consecutive_dl_tbfs()
@@ -817,7 +759,8 @@ static void test_2_consecutive_dl_tbfs()
 	trx->pdch[6].enable();
 	trx->pdch[7].enable();
 
-	dl_tbf1 = tbf_alloc_dl_tbf(bts, NULL, 0, ms_class, egprs_ms_class, 0);
+	dl_tbf1 = tbf_alloc_dl_tbf(bts, NULL, 0, ms_class, egprs_ms_class,
+				   false);
 	OSMO_ASSERT(dl_tbf1);
 
 	for (int i = 0; i < 8; i++) {
@@ -827,7 +770,8 @@ static void test_2_consecutive_dl_tbfs()
 	OSMO_ASSERT(numTs1 == 4);
 	printf("TBF1: numTs(%d)\n", numTs1);
 
-	dl_tbf2 = tbf_alloc_dl_tbf(bts, NULL, 0, ms_class, egprs_ms_class, 0);
+	dl_tbf2 = tbf_alloc_dl_tbf(bts, NULL, 0, ms_class, egprs_ms_class,
+				   false);
 	OSMO_ASSERT(dl_tbf2);
 
 	for (int i = 0; i < 8; i++) {
@@ -853,7 +797,7 @@ int main(int argc, char **argv)
 		abort();
 
 	msgb_talloc_ctx_init(tall_pcu_ctx, 0);
-	osmo_init_logging(&gprs_log_info);
+	osmo_init_logging2(tall_pcu_ctx, &gprs_log_info);
 	log_set_use_color(osmo_stderr_target, 0);
 	log_set_print_filename(osmo_stderr_target, 0);
 	if (getenv("LOGL_DEBUG"))
@@ -861,8 +805,10 @@ int main(int argc, char **argv)
 
 	test_alloc_a();
 	test_alloc_b();
-	test_successive_allocation();
-	test_many_connections();
+	test_successive_allocations();
+	test_many_connections(alloc_algorithm_a, 160, "A");
+	test_many_connections(alloc_algorithm_b, 32, "B");
+	test_many_connections(alloc_algorithm_dynamic, 160, "dynamic");
 	test_2_consecutive_dl_tbfs();
 	return EXIT_SUCCESS;
 }
