@@ -188,15 +188,15 @@ static void test_EPDAN_decode_tree(void)
 
 int main(int argc, char **argv)
 {
-	osmo_init_logging(&gprs_log_info);
-	log_set_use_color(osmo_stderr_target, 0);
-	log_set_print_filename(osmo_stderr_target, 0);
-	log_parse_category_mask(osmo_stderr_target, "DRLCMACUL,1");
-
 	tall_pcu_ctx = talloc_named_const(NULL, 1, "moiji-mobile bitcompTest context");
 	if (!tall_pcu_ctx)
 		abort();
 
+	osmo_init_logging2(tall_pcu_ctx, &gprs_log_info);
+	log_set_use_color(osmo_stderr_target, 0);
+	log_set_print_filename(osmo_stderr_target, 0);
+	log_parse_category_mask(osmo_stderr_target, "DRLCMACUL,1");
+
 	test_EPDAN_decode_tree();
 
 	if (getenv("TALLOC_REPORT_FULL"))
