@@ -27,6 +27,8 @@
 #include "rlc.h"
 #include "llc.h"
 #include "bts.h"
+#include <gprs_rlcmac.h>
+
 extern "C" {
 #include "pcu_vty.h"
 
@@ -44,6 +46,7 @@ extern "C" {
 
 void *tall_pcu_ctx;
 int16_t spoof_mnc = 0, spoof_mcc = 0;
+bool spoof_mnc_3_digits = false;
 
 static void check_coding_scheme(GprsCodingScheme& cs, GprsCodingScheme::Mode mode)
 {
@@ -1399,7 +1402,7 @@ int main(int argc, char **argv)
 		abort();
 
 	msgb_talloc_ctx_init(tall_pcu_ctx, 0);
-	osmo_init_logging(&gprs_log_info);
+	osmo_init_logging2(tall_pcu_ctx, &gprs_log_info);
 	log_set_use_color(osmo_stderr_target, 0);
 	log_set_print_filename(osmo_stderr_target, 0);
 
