@@ -27,6 +27,7 @@
 #include "gprs_bssgp_pcu.h"
 #include "pcu_l1_if.h"
 #include "decoding.h"
+#include <gprs_rlcmac.h>
 
 extern "C" {
 #include "pcu_vty.h"
@@ -46,14 +47,15 @@ extern "C" {
 
 void *tall_pcu_ctx;
 int16_t spoof_mnc = 0, spoof_mcc = 0;
+bool spoof_mnc_3_digits = false;
 
 static void check_tbf(gprs_rlcmac_tbf *tbf)
 {
 	OSMO_ASSERT(tbf);
 	if (tbf->state_is(GPRS_RLCMAC_WAIT_RELEASE))
-		OSMO_ASSERT(tbf->T == 3191 || tbf->T == 3193);
+		OSMO_ASSERT(tbf->timers_pending(T3191) || tbf->timers_pending(T3193));
 	if (tbf->state_is(GPRS_RLCMAC_RELEASING))
-		OSMO_ASSERT(tbf->T != 0);
+		OSMO_ASSERT(tbf->timers_pending(T_MAX));
 }
 
 static void test_tbf_base()
@@ -83,14 +85,14 @@ static void test_tbf_tlli_update()
 	 */
 	gprs_rlcmac_tbf *dl_tbf = tbf_alloc_dl_tbf(the_bts.bts_data(),
 						NULL,
-						0, 0, 0, 0);
+						0, 0, 0, false);
 	OSMO_ASSERT(dl_tbf != NULL);
 	dl_tbf->update_ms(0x2342, GPRS_RLCMAC_DL_TBF);
 	dl_tbf->set_ta(4);
 
 	gprs_rlcmac_tbf *ul_tbf = tbf_alloc_ul_tbf(the_bts.bts_data(),
 						dl_tbf->ms(),
-						0, 0, 0, 0);
+						0, 0, 0, false);
 	OSMO_ASSERT(ul_tbf != NULL);
 	ul_tbf->update_ms(0x2342, GPRS_RLCMAC_UL_TBF);
 
@@ -137,7 +139,9 @@ static void test_tbf_tlli_update()
 
 static uint8_t llc_data[200];
 
-int pcu_sock_send(struct msgb *msg)
+/* override, requires '-Wl,--wrap=pcu_sock_send' */
+int __real_pcu_sock_send(struct msgb *msg);
+int __wrap_pcu_sock_send(struct msgb *msg)
 {
 	return 0;
 }
@@ -170,13 +174,14 @@ static gprs_rlcmac_dl_tbf *create_dl_tbf(BTS *the_bts, uint8_t ms_class,
 
 	tfi = the_bts->tfi_find_free(GPRS_RLCMAC_DL_TBF, &trx_no, -1);
 	OSMO_ASSERT(tfi >= 0);
-	dl_tbf = tbf_alloc_dl_tbf(bts, NULL, trx_no, ms_class, egprs_ms_class, 1);
+	dl_tbf = tbf_alloc_dl_tbf(bts, NULL, trx_no, ms_class, egprs_ms_class,
+				  true);
 	dl_tbf->set_ta(0);
 	check_tbf(dl_tbf);
 
 	/* "Establish" the DL TBF */
-	dl_tbf->dl_ass_state = GPRS_RLCMAC_DL_ASS_SEND_ASS;
-	dl_tbf->set_state(GPRS_RLCMAC_FLOW);
+	TBF_SET_ASS_STATE_DL(dl_tbf, GPRS_RLCMAC_DL_ASS_SEND_ASS);
+	TBF_SET_STATE(dl_tbf, GPRS_RLCMAC_FLOW);
 	dl_tbf->m_wait_confirm = 0;
 	check_tbf(dl_tbf);
 
@@ -267,7 +272,7 @@ static void test_tbf_final_ack(enum test_tbf_final_ack_mode test_mode)
 	/* Queue a final ACK */
 	memset(rbb, 0, sizeof(rbb));
 	/* Receive a final ACK */
-	dl_tbf->rcvd_dl_ack(1, 1, rbb);
+	dl_tbf->rcvd_dl_ack(true, 1, rbb);
 
 	/* Clean up and ensure tbfs are in the correct state */
 	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_WAIT_RELEASE));
@@ -276,7 +281,7 @@ static void test_tbf_final_ack(enum test_tbf_final_ack_mode test_mode)
 	OSMO_ASSERT(new_tbf != dl_tbf);
 	OSMO_ASSERT(new_tbf->tfi() == 1);
 	check_tbf(dl_tbf);
-	dl_tbf->dl_ass_state = GPRS_RLCMAC_DL_ASS_NONE;
+	TBF_SET_ASS_STATE_DL(dl_tbf, GPRS_RLCMAC_DL_ASS_NONE);
 	if (test_mode == TEST_MODE_REVERSE_FREE) {
 		GprsMs::Guard guard(ms);
 		tbf_free(new_tbf);
@@ -295,6 +300,13 @@ static void test_tbf_final_ack(enum test_tbf_final_ack_mode test_mode)
 	printf("=== end %s ===\n", __func__);
 }
 
+/* Receive an ACK */
+#define RCV_ACK(fin, tbf, rbb) do {					\
+		tbf->rcvd_dl_ack(fin, tbf->window()->v_s(), rbb);	\
+		if (!fin)						\
+			OSMO_ASSERT(tbf->window()->window_empty());	\
+	} while(0)
+
 static void test_tbf_delayed_release()
 {
 	BTS the_bts;
@@ -340,16 +352,13 @@ static void test_tbf_delayed_release()
 
 	/* ACK all blocks */
 	memset(rbb, 0xff, sizeof(rbb));
-	/* Receive an ACK */
-	dl_tbf->rcvd_dl_ack(0, dl_tbf->m_window.v_s(), rbb);
-	OSMO_ASSERT(dl_tbf->m_window.window_empty());
+
+	RCV_ACK(false, dl_tbf, rbb); /* Receive an ACK */
 
 	/* Force sending of a single block containing an LLC dummy command */
 	request_dl_rlc_block(dl_tbf, &fn);
 
-	/* Receive an ACK */
-	dl_tbf->rcvd_dl_ack(0, dl_tbf->m_window.v_s(), rbb);
-	OSMO_ASSERT(dl_tbf->m_window.window_empty());
+	RCV_ACK(false, dl_tbf, rbb); /* Receive an ACK */
 
 	/* Timeout (make sure fn % 52 remains valid) */
 	fn += 52 * ((msecs_to_frames(bts->dl_tbf_idle_msec + 100) + 51)/ 52);
@@ -357,12 +366,11 @@ static void test_tbf_delayed_release()
 
 	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_FINISHED));
 
-	/* Receive a final ACK */
-	dl_tbf->rcvd_dl_ack(1, dl_tbf->m_window.v_s(), rbb);
+	RCV_ACK(true, dl_tbf, rbb); /* Receive a final ACK */
 
 	/* Clean up and ensure tbfs are in the correct state */
 	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_WAIT_RELEASE));
-	dl_tbf->dl_ass_state = GPRS_RLCMAC_DL_ASS_NONE;
+	TBF_SET_ASS_STATE_DL(dl_tbf, GPRS_RLCMAC_DL_ASS_NONE);
 	check_tbf(dl_tbf);
 	tbf_free(dl_tbf);
 	printf("=== end %s ===\n", __func__);
@@ -442,7 +450,7 @@ static void test_tbf_exhaustion()
 	bts = the_bts.bts_data();
 	setup_bts(&the_bts, ts_no);
 	gprs_bssgp_create_and_connect(bts, 33001, 0, 33001,
-		1234, 1234, 1234, 1, 1, 0, 0, 0);
+		1234, 1234, 1234, 1, 1, false, 0, 0, 0);
 
 	for (i = 0; i < 1024; i++) {
 		uint32_t tlli = 0xc0000000 + i;
@@ -485,7 +493,7 @@ static void test_tbf_dl_llc_loss()
 	bts->ms_idle_sec = 10; /* keep the MS object */
 
 	gprs_bssgp_create_and_connect(bts, 33001, 0, 33001,
-		1234, 1234, 1234, 1, 1, 0, 0, 0);
+		1234, 1234, 1234, 1, 1, false, 0, 0, 0);
 
 	/* Handle LLC frame 1 */
 	memset(buf, 1, sizeof(buf));
@@ -1347,7 +1355,7 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_two_phase_puan_URBB_no_length(BTS *t
 	request_dl_rlc_block(ul_tbf, fn);
 
 	check_tbf(ul_tbf);
-	OSMO_ASSERT(ul_tbf->ul_ack_state == GPRS_RLCMAC_UL_ACK_NONE);
+	OSMO_ASSERT(ul_tbf->ul_ack_state_is(GPRS_RLCMAC_UL_ACK_NONE));
 
 	ms = the_bts->ms_by_tlli(tlli);
 	OSMO_ASSERT(ms != NULL);
@@ -1431,7 +1439,7 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_two_phase_puan_URBB_with_length(BTS
 	request_dl_rlc_block(ul_tbf, fn);
 
 	check_tbf(ul_tbf);
-	OSMO_ASSERT(ul_tbf->ul_ack_state == GPRS_RLCMAC_UL_ACK_NONE);
+	OSMO_ASSERT(ul_tbf->ul_ack_state_is(GPRS_RLCMAC_UL_ACK_NONE));
 
 	ms = the_bts->ms_by_tlli(tlli);
 	OSMO_ASSERT(ms != NULL);
@@ -1519,7 +1527,7 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_two_phase_puan_CRBB(BTS *the_bts,
 	request_dl_rlc_block(ul_tbf, fn);
 
 	check_tbf(ul_tbf);
-	OSMO_ASSERT(ul_tbf->ul_ack_state == GPRS_RLCMAC_UL_ACK_NONE);
+	OSMO_ASSERT(ul_tbf->ul_ack_state_is(GPRS_RLCMAC_UL_ACK_NONE));
 
 	ms = the_bts->ms_by_tlli(tlli);
 	OSMO_ASSERT(ms != NULL);
@@ -1659,6 +1667,13 @@ static void transmit_dl_data(BTS *the_bts, uint32_t tlli, uint32_t *fn,
 	}
 }
 
+static inline void print_ta_tlli(const gprs_rlcmac_ul_tbf *ul_tbf, bool print_ms)
+{
+	fprintf(stderr, "Got '%s', TA=%d\n", ul_tbf->name(), ul_tbf->ta());
+	if (print_ms)
+		fprintf(stderr, "Got MS: TLLI = 0x%08x, TA = %d\n", ul_tbf->ms()->tlli(), ul_tbf->ms()->ta());
+}
+
 static void test_tbf_single_phase()
 {
 	BTS the_bts;
@@ -1668,7 +1683,6 @@ static void test_tbf_single_phase()
 	const char *imsi = "0011223344";
 	uint16_t qta = 31;
 	gprs_rlcmac_ul_tbf *ul_tbf;
-	GprsMs *ms;
 
 	printf("=== start %s ===\n", __func__);
 
@@ -1676,10 +1690,7 @@ static void test_tbf_single_phase()
 
 	ul_tbf = establish_ul_tbf_single_phase(&the_bts, ts_no, tlli, &fn, qta);
 
-	ms = ul_tbf->ms();
-	fprintf(stderr, "Got '%s', TA=%d\n", ul_tbf->name(), ul_tbf->ta());
-	fprintf(stderr, "Got MS: TLLI = 0x%08x, TA = %d\n", ms->tlli(), ms->ta());
-
+	print_ta_tlli(ul_tbf, true);
 	send_dl_data(&the_bts, tlli, imsi, (const uint8_t *)"TEST", 4);
 
 	printf("=== end %s ===\n", __func__);
@@ -1697,7 +1708,6 @@ static void test_tbf_egprs_two_phase_puan(void)
 	gprs_rlcmac_bts *bts;
 	uint8_t egprs_ms_class = 1;
 	gprs_rlcmac_ul_tbf *ul_tbf;
-	GprsMs *ms;
 	uint8_t test_data[256];
 
 	printf("=== start %s ===\n", __func__);
@@ -1716,38 +1726,25 @@ static void test_tbf_egprs_two_phase_puan(void)
 	ul_tbf = establish_ul_tbf_two_phase_puan_URBB_no_length(&the_bts, ts_no, tlli, &fn,
 		qta, ms_class, egprs_ms_class, ul_tbf);
 
-	ms = ul_tbf->ms();
-	fprintf(stderr, "Got '%s', TA=%d\n", ul_tbf->name(), ul_tbf->ta());
-	fprintf(stderr,
-		"Got MS: TLLI = 0x%08x, TA = %d\n", ms->tlli(), ms->ta());
+	print_ta_tlli(ul_tbf, true);
 	send_dl_data(&the_bts, tlli, imsi, test_data, sizeof(test_data));
 
-	ul_tbf->m_window.set_v_r(0);
-	ul_tbf->m_window.set_v_q(0);
+	ul_tbf->window()->reset_state();
 	/* Function to generate URBB with length */
 	ul_tbf = establish_ul_tbf_two_phase_puan_URBB_with_length(&the_bts, ts_no, tlli, &fn,
 		qta, ms_class, egprs_ms_class, ul_tbf);
 
-	ms = ul_tbf->ms();
-	fprintf(stderr, "Got '%s', TA=%d\n", ul_tbf->name(), ul_tbf->ta());
-	fprintf(stderr,
-		"Got MS: TLLI = 0x%08x, TA = %d\n", ms->tlli(), ms->ta());
-
+	print_ta_tlli(ul_tbf, true);
 	send_dl_data(&the_bts, tlli, imsi, test_data, sizeof(test_data));
 
-	ul_tbf->m_window.set_v_r(0);
-	ul_tbf->m_window.set_v_q(0);
+	ul_tbf->window()->reset_state();
 	/* Function to generate CRBB */
 	bts->ws_base = 128;
 	bts->ws_pdch = 64;
 	ul_tbf = establish_ul_tbf_two_phase_puan_CRBB(&the_bts, ts_no, tlli, &fn,
 		qta, ms_class, egprs_ms_class);
 
-	ms = ul_tbf->ms();
-	fprintf(stderr, "Got '%s', TA=%d\n", ul_tbf->name(), ul_tbf->ta());
-	fprintf(stderr,
-		"Got MS: TLLI = 0x%08x, TA = %d\n", ms->tlli(), ms->ta());
-
+	print_ta_tlli(ul_tbf, true);
 	send_dl_data(&the_bts, tlli, imsi, test_data, sizeof(test_data));
 
 	printf("=== end %s ===\n", __func__);
@@ -1844,7 +1841,6 @@ static void test_tbf_two_phase()
 	const char *imsi = "0011223344";
 	uint8_t ms_class = 1;
 	gprs_rlcmac_ul_tbf *ul_tbf;
-	GprsMs *ms;
 
 	printf("=== start %s ===\n", __func__);
 
@@ -1853,15 +1849,18 @@ static void test_tbf_two_phase()
 	ul_tbf = establish_ul_tbf_two_phase(&the_bts, ts_no, tlli, &fn, qta,
 		ms_class, 0);
 
-	ms = ul_tbf->ms();
-	fprintf(stderr, "Got '%s', TA=%d\n", ul_tbf->name(), ul_tbf->ta());
-	fprintf(stderr, "Got MS: TLLI = 0x%08x, TA = %d\n", ms->tlli(), ms->ta());
-
+	print_ta_tlli(ul_tbf, true);
 	send_dl_data(&the_bts, tlli, imsi, (const uint8_t *)"TEST", 4);
 
 	printf("=== end %s ===\n", __func__);
 }
 
+static inline void print_ms(const GprsMs *ms, bool old)
+{
+	fprintf(stderr, "%s MS: TLLI = 0x%08x, TA = %d, IMSI = %s, LLC = %zu\n",
+		old ? "Old" : "New", ms->tlli(), ms->ta(), ms->imsi(), ms->llc_queue()->size());
+}
+
 static void test_tbf_ra_update_rach()
 {
 	BTS the_bts;
@@ -1883,11 +1882,10 @@ static void test_tbf_ra_update_rach()
 		ms_class, 0);
 
 	ms1 = ul_tbf->ms();
-	fprintf(stderr, "Got '%s', TA=%d\n", ul_tbf->name(), ul_tbf->ta());
+	print_ta_tlli(ul_tbf, false);
 
 	send_dl_data(&the_bts, tlli1, imsi, (const uint8_t *)"RAU_ACCEPT", 10);
-	fprintf(stderr, "Old MS: TLLI = 0x%08x, TA = %d, IMSI = %s, LLC = %zu\n",
-		ms1->tlli(), ms1->ta(), ms1->imsi(), ms1->llc_queue()->size());
+	print_ms(ms1, true);
 
 	/* Send Packet Downlink Assignment to MS */
 	request_dl_rlc_block(ul_tbf, &fn);
@@ -1908,9 +1906,7 @@ static void test_tbf_ra_update_rach()
 
 	/* The PCU cannot know yet, that both TBF belong to the same MS */
 	OSMO_ASSERT(ms1 != ms2);
-
-	fprintf(stderr, "Old MS: TLLI = 0x%08x, TA = %d, IMSI = %s, LLC = %zu\n",
-		ms1->tlli(), ms1->ta(), ms1->imsi(), ms1->llc_queue()->size());
+	print_ms(ms1, true);
 
 	/* Send some downlink data along with the new TLLI and the IMSI so that
 	 * the PCU can see, that both MS objects belong to same MS */
@@ -1919,8 +1915,7 @@ static void test_tbf_ra_update_rach()
 	ms = the_bts.ms_by_imsi(imsi);
 	OSMO_ASSERT(ms == ms2);
 
-	fprintf(stderr, "New MS: TLLI = 0x%08x, TA = %d, IMSI = %s, LLC = %zu\n",
-		ms2->tlli(), ms2->ta(), ms2->imsi(), ms2->llc_queue()->size());
+	print_ms(ms2, false);
 
 	ms = the_bts.ms_by_tlli(tlli1);
 	OSMO_ASSERT(ms == NULL);
@@ -1951,12 +1946,11 @@ static void test_tbf_dl_flow_and_rach_two_phase()
 		ms_class, 0);
 
 	ms1 = ul_tbf->ms();
-	fprintf(stderr, "Got '%s', TA=%d\n", ul_tbf->name(), ul_tbf->ta());
+	print_ta_tlli(ul_tbf, false);
 
 	send_dl_data(&the_bts, tlli1, imsi, (const uint8_t *)"DATA 1 *************", 20);
 	send_dl_data(&the_bts, tlli1, imsi, (const uint8_t *)"DATA 2 *************", 20);
-	fprintf(stderr, "Old MS: TLLI = 0x%08x, TA = %d, IMSI = %s, LLC = %zu\n",
-		ms1->tlli(), ms1->ta(), ms1->imsi(), ms1->llc_queue()->size());
+	print_ms(ms1, true);
 
 	OSMO_ASSERT(ms1->llc_queue()->size() == 2);
 	dl_tbf = ms1->dl_tbf();
@@ -1972,8 +1966,7 @@ static void test_tbf_dl_flow_and_rach_two_phase()
 		ms_class, 0);
 
 	ms2 = ul_tbf->ms();
-	fprintf(stderr, "New MS: TLLI = 0x%08x, TA = %d, IMSI = %s, LLC = %zu\n",
-		ms2->tlli(), ms2->ta(), ms2->imsi(), ms2->llc_queue()->size());
+	print_ms(ms2, false);
 
 	/* This should be the same MS object */
 	OSMO_ASSERT(ms2 == ms1);
@@ -2012,12 +2005,11 @@ static void test_tbf_dl_flow_and_rach_single_phase()
 		ms_class, 0);
 
 	ms1 = ul_tbf->ms();
-	fprintf(stderr, "Got '%s', TA=%d\n", ul_tbf->name(), ul_tbf->ta());
+	print_ta_tlli(ul_tbf, false);
 
 	send_dl_data(&the_bts, tlli1, imsi, (const uint8_t *)"DATA 1 *************", 20);
 	send_dl_data(&the_bts, tlli1, imsi, (const uint8_t *)"DATA 2 *************", 20);
-	fprintf(stderr, "Old MS: TLLI = 0x%08x, TA = %d, IMSI = %s, LLC = %zu\n",
-		ms1->tlli(), ms1->ta(), ms1->imsi(), ms1->llc_queue()->size());
+	print_ms(ms1, true);
 
 	OSMO_ASSERT(ms1->llc_queue()->size() == 2);
 	dl_tbf = ms1->dl_tbf();
@@ -2032,8 +2024,7 @@ static void test_tbf_dl_flow_and_rach_single_phase()
 	ul_tbf = establish_ul_tbf_single_phase(&the_bts, ts_no, tlli1, &fn, qta);
 
 	ms2 = ul_tbf->ms();
-	fprintf(stderr, "New MS: TLLI = 0x%08x, TA = %d, IMSI = %s, LLC = %zu\n",
-		ms2->tlli(), ms2->ta(), ms2->imsi(), ms2->llc_queue()->size());
+	print_ms(ms2, false);
 
 	/* There should be a different MS object */
 	OSMO_ASSERT(ms2 != ms1);
@@ -2074,7 +2065,7 @@ static void test_tbf_dl_reuse()
 		ms_class, 0);
 
 	ms1 = ul_tbf->ms();
-	fprintf(stderr, "Got '%s', TA=%d\n", ul_tbf->name(), ul_tbf->ta());
+	print_ta_tlli(ul_tbf, false);
 
 	/* Send some LLC frames */
 	for (i = 0; i < 40; i++) {
@@ -2087,8 +2078,7 @@ static void test_tbf_dl_reuse()
 		send_dl_data(&the_bts, tlli1, imsi, (const uint8_t *)buf, rc);
 	}
 
-	fprintf(stderr, "Old MS: TLLI = 0x%08x, TA = %d, IMSI = %s, LLC = %zu\n",
-		ms1->tlli(), ms1->ta(), ms1->imsi(), ms1->llc_queue()->size());
+	print_ms(ms1, true);
 
 	/* Send Packet Downlink Assignment to MS */
 	request_dl_rlc_block(ul_tbf, &fn);
@@ -2172,7 +2162,7 @@ static void test_tbf_gprs_egprs()
 	bts->egprs_enabled = 1;
 
 	gprs_bssgp_create_and_connect(bts, 33001, 0, 33001,
-		1234, 1234, 1234, 1, 1, 0, 0, 0);
+		1234, 1234, 1234, 1, 1, false, 0, 0, 0);
 
 	/* Does not support EGPRS */
 	rc = gprs_rlcmac_dl_tbf::handle(bts, tlli, 0, imsi, ms_class, 0,
@@ -2184,6 +2174,34 @@ static void test_tbf_gprs_egprs()
 	gprs_bssgp_destroy();
 }
 
+static inline void ws_check(gprs_rlcmac_dl_tbf *dl_tbf, const char *test, uint8_t exp_slots, uint16_t exp_ws,
+			    bool free, bool end)
+{
+	if (!dl_tbf) {
+		fprintf(stderr, "%s(): FAILED (NULL TBF)\n", test);
+		return;
+	}
+
+	fprintf(stderr, "DL TBF slots: 0x%02x, N: %d, WS: %d",
+		dl_tbf->dl_slots(),
+		pcu_bitcount(dl_tbf->dl_slots()),
+		dl_tbf->window_size());
+
+	if (pcu_bitcount(dl_tbf->dl_slots()) != exp_slots || dl_tbf->window_size() != exp_ws)
+		fprintf(stderr, "%s(): DL TBF FAILED: dl_slots = %u (exp. %u), WS = %u (exp. %u)",
+			test, pcu_bitcount(dl_tbf->dl_slots()), 4, dl_tbf->window_size(), 128 + 4 * 64);
+
+	fprintf(stderr, "\n");
+
+	if (free)
+		tbf_free(dl_tbf);
+
+	if (end) {
+		printf("=== end %s ===\n", test);
+		gprs_bssgp_destroy();
+	}
+}
+
 static void test_tbf_ws()
 {
 	BTS the_bts;
@@ -2206,37 +2224,20 @@ static void test_tbf_ws()
 	bts->trx[0].pdch[5].enable();
 
 	gprs_bssgp_create_and_connect(bts, 33001, 0, 33001,
-		1234, 1234, 1234, 1, 1, 0, 0, 0);
+		1234, 1234, 1234, 1, 1, false, 0, 0, 0);
 
 	/* Does no support EGPRS */
-	dl_tbf = tbf_alloc_dl_tbf(bts, NULL, 0, ms_class, 0, 0);
-	OSMO_ASSERT(dl_tbf != NULL);
-	fprintf(stderr, "DL TBF slots: 0x%02x, N: %d, WS: %d\n",
-		dl_tbf->dl_slots(),
-		pcu_bitcount(dl_tbf->dl_slots()),
-		dl_tbf->window()->ws());
-	OSMO_ASSERT(pcu_bitcount(dl_tbf->dl_slots()) == 4);
-	OSMO_ASSERT(dl_tbf->window()->ws() == 64);
-	tbf_free(dl_tbf);
+	dl_tbf = tbf_alloc_dl_tbf(bts, NULL, 0, ms_class, 0, false);
+
+	ws_check(dl_tbf, __func__, 4, 64, true, false);
 
 	/* EGPRS-only */
 	bts->egprs_enabled = 1;
 
 	/* Does support EGPRS */
-	dl_tbf = tbf_alloc_dl_tbf(bts, NULL, 0, ms_class, ms_class, 0);
+	dl_tbf = tbf_alloc_dl_tbf(bts, NULL, 0, ms_class, ms_class, false);
 
-	OSMO_ASSERT(dl_tbf != NULL);
-	fprintf(stderr, "DL TBF slots: 0x%02x, N: %d, WS: %d\n",
-		dl_tbf->dl_slots(),
-		pcu_bitcount(dl_tbf->dl_slots()),
-		dl_tbf->window()->ws());
-	OSMO_ASSERT(pcu_bitcount(dl_tbf->dl_slots()) == 4);
-	OSMO_ASSERT(dl_tbf->window()->ws() == 128 + 4 * 64);
-	tbf_free(dl_tbf);
-
-	printf("=== end %s ===\n", __func__);
-
-	gprs_bssgp_destroy();
+	ws_check(dl_tbf, __func__, 4, 128 + 4 * 64, true, true);
 }
 
 static void test_tbf_update_ws(void)
@@ -2261,38 +2262,20 @@ static void test_tbf_update_ws(void)
 	bts->trx[0].pdch[5].enable();
 
 	gprs_bssgp_create_and_connect(bts, 33001, 0, 33001,
-		1234, 1234, 1234, 1, 1, 0, 0, 0);
+		1234, 1234, 1234, 1, 1, false, 0, 0, 0);
 
 	/* EGPRS-only */
 	bts->egprs_enabled = 1;
 
 	/* Does support EGPRS */
-	dl_tbf = tbf_alloc_dl_tbf(bts, NULL, 0, ms_class, ms_class, 1);
+	dl_tbf = tbf_alloc_dl_tbf(bts, NULL, 0, ms_class, ms_class, true);
 
-	OSMO_ASSERT(dl_tbf != NULL);
-	fprintf(stderr, "DL TBF slots: 0x%02x, N: %d, WS: %d\n",
-		dl_tbf->dl_slots(),
-		pcu_bitcount(dl_tbf->dl_slots()),
-		dl_tbf->window()->ws());
-	OSMO_ASSERT(pcu_bitcount(dl_tbf->dl_slots()) == 1);
-	OSMO_ASSERT(dl_tbf->window()->ws() == 128 + 1 * 64);
+	ws_check(dl_tbf, __func__, 1, 128 + 1 * 64, false, false);
 
 	dl_tbf->update();
 
 	/* window size should be 384 */
-	OSMO_ASSERT(dl_tbf != NULL);
-	fprintf(stderr, "DL TBF slots: 0x%02x, N: %d, WS: %d\n",
-		dl_tbf->dl_slots(),
-		pcu_bitcount(dl_tbf->dl_slots()),
-		dl_tbf->window()->ws());
-	OSMO_ASSERT(pcu_bitcount(dl_tbf->dl_slots()) == 4);
-	OSMO_ASSERT(dl_tbf->window()->ws() == 128 + 4 * 64);
-
-	tbf_free(dl_tbf);
-
-	printf("=== end %s ===\n", __func__);
-
-	gprs_bssgp_destroy();
+	ws_check(dl_tbf, __func__, 4, 128 + 4 * 64, true, true);
 }
 
 static void test_tbf_puan_urbb_len(void)
@@ -2306,7 +2289,6 @@ static void test_tbf_puan_urbb_len(void)
 	uint8_t ms_class = 1;
 	uint8_t egprs_ms_class = 1;
 	gprs_rlcmac_ul_tbf *ul_tbf;
-	GprsMs *ms;
 	uint8_t test_data[256];
 
 	printf("=== start %s ===\n", __func__);
@@ -2320,11 +2302,7 @@ static void test_tbf_puan_urbb_len(void)
 	ul_tbf = puan_urbb_len_issue(&the_bts, ts_no, tlli, &fn, qta,
 		ms_class, egprs_ms_class);
 
-	ms = ul_tbf->ms();
-	fprintf(stderr, "Got '%s', TA=%d\n", ul_tbf->name(), ul_tbf->ta());
-	fprintf(stderr,
-		"Got MS: TLLI = 0x%08x, TA = %d\n", ms->tlli(), ms->ta());
-
+	print_ta_tlli(ul_tbf, true);
 	send_dl_data(&the_bts, tlli, imsi, test_data, sizeof(test_data));
 
 	printf("=== end %s ===\n", __func__);
@@ -2456,7 +2434,6 @@ static void test_tbf_li_decoding(void)
 	uint8_t ms_class = 1;
 	uint8_t egprs_ms_class = 1;
 	gprs_rlcmac_ul_tbf *ul_tbf;
-	GprsMs *ms;
 	uint8_t test_data[256];
 
 	printf("=== start %s ===\n", __func__);
@@ -2470,11 +2447,7 @@ static void test_tbf_li_decoding(void)
 	ul_tbf = tbf_li_decoding(&the_bts, ts_no, tlli, &fn, qta,
 		ms_class, egprs_ms_class);
 
-	ms = ul_tbf->ms();
-	fprintf(stderr, "Got '%s', TA=%d\n", ul_tbf->name(), ul_tbf->ta());
-	fprintf(stderr,
-		"Got MS: TLLI = 0x%08x, TA = %d\n", ms->tlli(), ms->ta());
-
+	print_ta_tlli(ul_tbf, true);
 	send_dl_data(&the_bts, tlli, imsi, test_data, sizeof(test_data));
 
 	printf("=== end %s ===\n", __func__);
@@ -2528,7 +2501,7 @@ static void test_tbf_epdan_out_of_rx_window(void)
 
 	dl_tbf = create_dl_tbf(&the_bts, ms_class, egprs_ms_class, &trx_no);
 	dl_tbf->update_ms(tlli, GPRS_RLCMAC_DL_TBF);
-	prlcdlwindow = &dl_tbf->m_window;
+	prlcdlwindow = dl_tbf->window();
 	prlcmvb = &prlcdlwindow->m_v_b;
 	prlcdlwindow->m_v_s = 1288;
 	prlcdlwindow->m_v_a = 1176;
@@ -2560,7 +2533,7 @@ static void test_tbf_epdan_out_of_rx_window(void)
 
 	Decoding::decode_egprs_acknack_bits(
 		&ack_nack->EGPRS_AckNack.Desc, &bits,
-		&bsn_begin, &bsn_end, &dl_tbf->m_window);
+		&bsn_begin, &bsn_end, dl_tbf->window());
 
 	dl_tbf->rcvd_dl_ack(
 		ack_nack->EGPRS_AckNack.Desc.FINAL_ACK_INDICATION,
@@ -2587,7 +2560,6 @@ static void test_tbf_egprs_two_phase_spb(void)
 	uint8_t ms_class = 1;
 	uint8_t egprs_ms_class = 1;
 	gprs_rlcmac_ul_tbf *ul_tbf;
-	GprsMs *ms;
 	uint8_t test_data[256];
 
 	printf("=== start %s ===\n", __func__);
@@ -2601,11 +2573,7 @@ static void test_tbf_egprs_two_phase_spb(void)
 	ul_tbf = establish_ul_tbf_two_phase_spb(&the_bts, ts_no, tlli, &fn, qta,
 		ms_class, egprs_ms_class);
 
-	ms = ul_tbf->ms();
-	fprintf(stderr, "Got '%s', TA=%d\n", ul_tbf->name(), ul_tbf->ta());
-	fprintf(stderr,
-		"Got MS: TLLI = 0x%08x, TA = %d\n", ms->tlli(), ms->ta());
-
+	print_ta_tlli(ul_tbf, true);
 	send_dl_data(&the_bts, tlli, imsi, test_data, sizeof(test_data));
 
 	printf("=== end %s ===\n", __func__);
@@ -2622,7 +2590,6 @@ static void test_tbf_egprs_two_phase()
 	uint8_t ms_class = 1;
 	uint8_t egprs_ms_class = 1;
 	gprs_rlcmac_ul_tbf *ul_tbf;
-	GprsMs *ms;
 	uint8_t test_data[256];
 
 	printf("=== start %s ===\n", __func__);
@@ -2636,10 +2603,7 @@ static void test_tbf_egprs_two_phase()
 	ul_tbf = establish_ul_tbf_two_phase(&the_bts, ts_no, tlli, &fn, qta,
 		ms_class, egprs_ms_class);
 
-	ms = ul_tbf->ms();
-	fprintf(stderr, "Got '%s', TA=%d\n", ul_tbf->name(), ul_tbf->ta());
-	fprintf(stderr, "Got MS: TLLI = 0x%08x, TA = %d\n", ms->tlli(), ms->ta());
-
+	print_ta_tlli(ul_tbf, true);
 	send_dl_data(&the_bts, tlli, imsi, test_data, sizeof(test_data));
 
 	printf("=== end %s ===\n", __func__);
@@ -2694,12 +2658,11 @@ static void establish_and_use_egprs_dl_tbf(BTS *the_bts, int mcs)
 
 	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_FLOW));
 
-	/* Receive a final ACK */
-	dl_tbf->rcvd_dl_ack(1, dl_tbf->m_window.v_s(), rbb);
+	RCV_ACK(true, dl_tbf, rbb); /* Receive a final ACK */
 
 	/* Clean up and ensure tbfs are in the correct state */
 	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_WAIT_RELEASE));
-	dl_tbf->dl_ass_state = GPRS_RLCMAC_DL_ASS_NONE;
+	TBF_SET_ASS_STATE_DL(dl_tbf, GPRS_RLCMAC_DL_ASS_NONE);
 	check_tbf(dl_tbf);
 	tbf_free(dl_tbf);
 }
@@ -2744,17 +2707,40 @@ static void tbf_cleanup(gprs_rlcmac_dl_tbf *dl_tbf)
 {
 	uint8_t rbb[64/8];
 
-	/* Receive a final ACK */
-	dl_tbf->rcvd_dl_ack(1, dl_tbf->m_window.v_s(), rbb);
+	RCV_ACK(true, dl_tbf, rbb); /* Receive a final ACK */
 
 	/* Clean up and ensure tbfs are in the correct state */
 	OSMO_ASSERT(dl_tbf->state_is(GPRS_RLCMAC_WAIT_RELEASE));
-	dl_tbf->dl_ass_state = GPRS_RLCMAC_DL_ASS_NONE;
+	TBF_SET_ASS_STATE_DL(dl_tbf, GPRS_RLCMAC_DL_ASS_NONE);
 	check_tbf(dl_tbf);
 	tbf_free(dl_tbf);
 
 }
 
+#define NACK(tbf, x) do {					\
+		tbf->window()->m_v_b.mark_nacked(x);		\
+		OSMO_ASSERT(tbf->window()->m_v_b.is_nacked(x));	\
+	} while(0)
+
+#define CHECK_UNACKED(tbf, cs, bsn) do {				             \
+		OSMO_ASSERT(tbf->window()->m_v_b.is_unacked(bsn));	             \
+		OSMO_ASSERT(tbf->m_rlc.block(bsn)->cs_current_trans.to_num() == cs); \
+	} while(0)
+
+#define CHECK_NACKED(tbf, cs, bsn) do {					             \
+		OSMO_ASSERT(tbf->window()->m_v_b.is_nacked(bsn));	             \
+		OSMO_ASSERT(tbf->m_rlc.block(bsn)->cs_current_trans.to_num() == cs); \
+	} while(0)
+
+#define MAKE_ACKED(m, tbf, fn, cs, check_unacked) do {			\
+		m = tbf->create_dl_acked_block(fn, tbf->control_ts);	\
+		OSMO_ASSERT(m);						\
+		if (check_unacked)					\
+			CHECK_UNACKED(tbf, cs, 0);			\
+		else							\
+			CHECK_NACKED(tbf, cs, 0);			\
+	} while(0)
+
 static void egprs_spb_to_normal_validation(BTS *the_bts,
 		unsigned int mcs, unsigned int demanded_mcs)
 {
@@ -2779,16 +2765,13 @@ static void egprs_spb_to_normal_validation(BTS *the_bts,
 
 	fn = fn_add_blocks(fn, 1);
 	/* Send first RLC data block BSN 0 */
-	msg = dl_tbf->create_dl_acked_block(fn, dl_tbf->control_ts);
-	OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(0));
-	OSMO_ASSERT(dl_tbf->m_rlc.block(0)->cs_current_trans.to_num()
-			== mcs);
+	MAKE_ACKED(msg, dl_tbf, fn, mcs, true);
 
 	egprs2 = (struct gprs_rlc_dl_header_egprs_2 *) msg->data;
-	bsn1 = (egprs2->bsn1_hi << 9) || (egprs2->bsn1_mid << 1)
-			|| (egprs2->bsn1_lo);
-	dl_tbf->m_window.m_v_b.mark_nacked(0);
-	OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_nacked(0));
+	bsn1 = (egprs2->bsn1_hi << 9) | (egprs2->bsn1_mid << 1) | (egprs2->bsn1_lo);
+
+	NACK(dl_tbf, 0);
+
 	OSMO_ASSERT(bsn1 == 0);
 
 	dl_tbf->ms()->set_current_cs_dl
@@ -2798,10 +2781,7 @@ static void egprs_spb_to_normal_validation(BTS *the_bts,
 	fn = fn_add_blocks(fn, 1);
 
 	/* Send first segment with demanded_mcs */
-	msg = dl_tbf->create_dl_acked_block(fn, dl_tbf->control_ts);
-	OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_nacked(0));
-	OSMO_ASSERT(dl_tbf->m_rlc.block(0)->cs_current_trans.to_num()
-			== demanded_mcs);
+	MAKE_ACKED(msg, dl_tbf, fn, demanded_mcs, false);
 	OSMO_ASSERT(dl_tbf->m_rlc.block(0)->spb_status.block_status_dl
 			== EGPRS_RESEG_FIRST_SEG_SENT);
 
@@ -2812,18 +2792,14 @@ static void egprs_spb_to_normal_validation(BTS *the_bts,
 	OSMO_ASSERT(egprs3->cps == 3);
 
 	/* Send second segment with demanded_mcs */
-	msg = dl_tbf->create_dl_acked_block(fn, dl_tbf->control_ts);
-	OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(0));
-	OSMO_ASSERT(dl_tbf->m_rlc.block(0)->cs_current_trans.to_num()
-			== demanded_mcs);
+	MAKE_ACKED(msg, dl_tbf, fn, demanded_mcs, true);
 	OSMO_ASSERT(dl_tbf->m_rlc.block(0)->spb_status.block_status_dl
 			== EGPRS_RESEG_SECOND_SEG_SENT);
 
 	egprs3 = (struct gprs_rlc_dl_header_egprs_3 *) msg->data;
 	/* Table 10.4.8a.3.1 of 44.060 */
 	OSMO_ASSERT(egprs3->spb == 3);
-	bsn2 = (egprs3->bsn1_hi << 9) || (egprs3->bsn1_mid << 1) ||
-			(egprs3->bsn1_lo);
+	bsn2 = (egprs3->bsn1_hi << 9) | (egprs3->bsn1_mid << 1) | (egprs3->bsn1_lo);
 	OSMO_ASSERT(bsn2 == bsn1);
 
 	/* Table 10.4.8a.3.1 of 44.060 */
@@ -2834,15 +2810,14 @@ static void egprs_spb_to_normal_validation(BTS *the_bts,
 		(static_cast < GprsCodingScheme::Scheme >
 			(GprsCodingScheme::CS4 + mcs));
 
-	dl_tbf->m_window.m_v_b.mark_nacked(0);
-	OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_nacked(0));
+	NACK(dl_tbf, 0);
+
 	msg = dl_tbf->create_dl_acked_block(fn, dl_tbf->control_ts);
 	egprs2 = (struct gprs_rlc_dl_header_egprs_2 *) msg->data;
 
 	/* Table 10.4.8a.3.1 of 44.060 */
 	OSMO_ASSERT(egprs2->cps == 0);
-	bsn3 = (egprs2->bsn1_hi << 9) || (egprs2->bsn1_mid << 1) ||
-			(egprs2->bsn1_lo);
+	bsn3 = (egprs2->bsn1_hi << 9) | (egprs2->bsn1_mid << 1) | (egprs2->bsn1_lo);
 	OSMO_ASSERT(bsn3 == bsn2);
 
 	tbf_cleanup(dl_tbf);
@@ -2876,13 +2851,9 @@ static void establish_and_use_egprs_dl_tbf_for_spb(BTS *the_bts,
 
 	fn = fn_add_blocks(fn, 1);
 	/* Send first RLC data block BSN 0 */
-	msg = dl_tbf->create_dl_acked_block(fn, dl_tbf->control_ts);
-	OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(0));
-	OSMO_ASSERT(dl_tbf->m_rlc.block(0)->cs_current_trans.to_num()
-			== mcs);
+	MAKE_ACKED(msg, dl_tbf, fn, mcs, true);
 
-	dl_tbf->m_window.m_v_b.mark_nacked(0);
-	OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_nacked(0));
+	NACK(dl_tbf, 0);
 
 	dl_tbf->ms()->set_current_cs_dl
 		(static_cast < GprsCodingScheme::Scheme >
@@ -2891,10 +2862,7 @@ static void establish_and_use_egprs_dl_tbf_for_spb(BTS *the_bts,
 	fn = fn_add_blocks(fn, 1);
 
 	/* Send first segment with demanded_mcs */
-	msg = dl_tbf->create_dl_acked_block(fn, dl_tbf->control_ts);
-	OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_nacked(0));
-	OSMO_ASSERT(dl_tbf->m_rlc.block(0)->cs_current_trans.to_num()
-			== demanded_mcs);
+	MAKE_ACKED(msg, dl_tbf, fn, demanded_mcs, false);
 	OSMO_ASSERT(dl_tbf->m_rlc.block(0)->spb_status.block_status_dl
 			== EGPRS_RESEG_FIRST_SEG_SENT);
 
@@ -2918,10 +2886,7 @@ static void establish_and_use_egprs_dl_tbf_for_spb(BTS *the_bts,
 	}
 
 	/* Send second segment with demanded_mcs */
-	msg = dl_tbf->create_dl_acked_block(fn, dl_tbf->control_ts);
-	OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(0));
-	OSMO_ASSERT(dl_tbf->m_rlc.block(0)->cs_current_trans.to_num()
-			== demanded_mcs);
+	MAKE_ACKED(msg, dl_tbf, fn, demanded_mcs, true);
 	OSMO_ASSERT(dl_tbf->m_rlc.block(0)->spb_status.block_status_dl
 			== EGPRS_RESEG_SECOND_SEG_SENT);
 
@@ -2970,19 +2935,11 @@ static void establish_and_use_egprs_dl_tbf_for_retx(BTS *the_bts,
 		((mcs == 7) && (demanded_mcs < 7))) {
 		fn = fn_add_blocks(fn, 1);
 		/* Send 2 RLC data block */
-		msg = dl_tbf->create_dl_acked_block(fn, dl_tbf->control_ts);
-		OSMO_ASSERT(msg);
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(0));
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(1));
-		OSMO_ASSERT(dl_tbf->m_rlc.block(0)->cs_current_trans.to_num()
-				== mcs);
-		OSMO_ASSERT(dl_tbf->m_rlc.block(1)->cs_current_trans.to_num()
-				== mcs);
+		MAKE_ACKED(msg, dl_tbf, fn, mcs, true);
+		CHECK_UNACKED(dl_tbf, mcs, 1);
 
-		dl_tbf->m_window.m_v_b.mark_nacked(0);
-		dl_tbf->m_window.m_v_b.mark_nacked(1);
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_nacked(0));
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_nacked(1));
+		NACK(dl_tbf, 0);
+		NACK(dl_tbf, 1);
 
 		/* Set the demanded MCS to demanded_mcs */
 		dl_tbf->ms()->set_current_cs_dl
@@ -2991,43 +2948,26 @@ static void establish_and_use_egprs_dl_tbf_for_retx(BTS *the_bts,
 
 		fn = fn_add_blocks(fn, 1);
 		/* Retransmit the first RLC data block with demanded_mcs */
-		msg = dl_tbf->create_dl_acked_block(fn, dl_tbf->control_ts);
-		OSMO_ASSERT(msg);
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(0));
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_nacked(1));
-		OSMO_ASSERT(dl_tbf->m_rlc.block(0)->cs_current_trans.to_num()
-				== demanded_mcs);
+		MAKE_ACKED(msg, dl_tbf, fn, demanded_mcs, true);
+		CHECK_NACKED(dl_tbf, mcs, 1);
 
 		fn = fn_add_blocks(fn, 1);
 		/* Retransmit the second RLC data block with demanded_mcs */
-		msg = dl_tbf->create_dl_acked_block(fn, dl_tbf->control_ts);
-		OSMO_ASSERT(msg);
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(0));
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(1));
-		OSMO_ASSERT(dl_tbf->m_rlc.block(1)->cs_current_trans.to_num()
-				== demanded_mcs);
+		MAKE_ACKED(msg, dl_tbf, fn, demanded_mcs, true);
+		CHECK_UNACKED(dl_tbf, demanded_mcs, 1);
 	} else if (((mcs == 5) && (demanded_mcs > 6)) ||
 		((mcs == 6) && (demanded_mcs > 8))) {
 		fn = fn_add_blocks(fn, 1);
 		/* Send first RLC data block BSN 0 */
-		msg = dl_tbf->create_dl_acked_block(fn, dl_tbf->control_ts);
-		OSMO_ASSERT(msg);
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(0));
-		OSMO_ASSERT(dl_tbf->m_rlc.block(0)->cs_current_trans.to_num()
-				== mcs);
+		MAKE_ACKED(msg, dl_tbf, fn, mcs, true);
 
 		fn = fn_add_blocks(fn, 1);
 		/* Send second RLC data block BSN 1 */
-		msg = dl_tbf->create_dl_acked_block(fn, dl_tbf->control_ts);
-		OSMO_ASSERT(msg);
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(1));
-		OSMO_ASSERT(dl_tbf->m_rlc.block(1)->cs_current_trans.to_num()
-				== mcs);
+		MAKE_ACKED(msg, dl_tbf, fn, mcs, true);
+		CHECK_UNACKED(dl_tbf, mcs, 1);
 
-		dl_tbf->m_window.m_v_b.mark_nacked(0);
-		dl_tbf->m_window.m_v_b.mark_nacked(1);
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_nacked(0));
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_nacked(1));
+		NACK(dl_tbf, 0);
+		NACK(dl_tbf, 1);
 
 		dl_tbf->ms()->set_current_cs_dl
 			(static_cast < GprsCodingScheme::Scheme >
@@ -3035,63 +2975,34 @@ static void establish_and_use_egprs_dl_tbf_for_retx(BTS *the_bts,
 
 		fn = fn_add_blocks(fn, 1);
 		/* Send first, second RLC data blocks with demanded_mcs */
-		msg = dl_tbf->create_dl_acked_block(fn, dl_tbf->control_ts);
-		OSMO_ASSERT(msg);
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(0));
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(1));
-		OSMO_ASSERT(dl_tbf->m_rlc.block(0)->cs_current_trans.to_num()
-				== demanded_mcs);
-		OSMO_ASSERT(dl_tbf->m_rlc.block(1)->cs_current_trans.to_num()
-				== demanded_mcs);
+		MAKE_ACKED(msg, dl_tbf, fn, demanded_mcs, true);
+		CHECK_UNACKED(dl_tbf, demanded_mcs, 1);
 	} else if (mcs > 6) {
 		/* No Mcs change cases are handled here for mcs > MCS6*/
 		fn = fn_add_blocks(fn, 1);
 		/* Send first,second RLC data blocks */
-		msg = dl_tbf->create_dl_acked_block(fn, dl_tbf->control_ts);
-		OSMO_ASSERT(msg);
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(0));
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(1));
-		OSMO_ASSERT(dl_tbf->m_rlc.block(0)->cs_current_trans.to_num()
-				== mcs);
-		OSMO_ASSERT(dl_tbf->m_rlc.block(1)->cs_current_trans.to_num()
-				== mcs);
+		MAKE_ACKED(msg, dl_tbf, fn, mcs, true);
+		CHECK_UNACKED(dl_tbf, mcs, 1);
 
-		dl_tbf->m_window.m_v_b.mark_nacked(0);
-		dl_tbf->m_window.m_v_b.mark_nacked(1);
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_nacked(0));
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_nacked(1));
+		NACK(dl_tbf, 0);
+		NACK(dl_tbf, 1);
 
 		fn = fn_add_blocks(fn, 1);
 		/* Send first,second RLC data blocks with demanded_mcs*/
-		msg = dl_tbf->create_dl_acked_block(fn, dl_tbf->control_ts);
-		OSMO_ASSERT(msg);
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(0));
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(1));
-		OSMO_ASSERT(dl_tbf->m_rlc.block(0)->cs_current_trans.to_num()
-				== mcs);
-		OSMO_ASSERT(dl_tbf->m_rlc.block(1)->cs_current_trans.to_num()
-				== mcs);
+		MAKE_ACKED(msg, dl_tbf, fn, mcs, true);
+		CHECK_UNACKED(dl_tbf, mcs, 1);
 	} else {
 
 		/* No MCS change cases are handled here for mcs <= MCS6*/
 		fn = fn_add_blocks(fn, 1);
 		/* Send first RLC data block */
-		msg = dl_tbf->create_dl_acked_block(fn, dl_tbf->control_ts);
-		OSMO_ASSERT(msg);
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(0));
-		OSMO_ASSERT(dl_tbf->m_rlc.block(0)->cs_current_trans.to_num()
-				== mcs);
+		MAKE_ACKED(msg, dl_tbf, fn, mcs, true);
 
-		dl_tbf->m_window.m_v_b.mark_nacked(0);
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_nacked(0));
+		NACK(dl_tbf, 0);
 
 		fn = fn_add_blocks(fn, 1);
 		/* Send first RLC data block with demanded_mcs */
-		msg = dl_tbf->create_dl_acked_block(fn, dl_tbf->control_ts);
-		OSMO_ASSERT(msg);
-		OSMO_ASSERT(dl_tbf->m_window.m_v_b.is_unacked(0));
-		OSMO_ASSERT(dl_tbf->m_rlc.block(0)->cs_current_trans.to_num()
-				== mcs);
+		MAKE_ACKED(msg, dl_tbf, fn, mcs, true);
 	}
 
 	tbf_cleanup(dl_tbf);
@@ -3322,12 +3233,13 @@ int main(int argc, char **argv)
 		abort();
 
 	msgb_talloc_ctx_init(tall_pcu_ctx, 0);
-	osmo_init_logging(&gprs_log_info);
+	osmo_init_logging2(tall_pcu_ctx, &gprs_log_info);
 	log_set_use_color(osmo_stderr_target, 0);
 	log_set_print_filename(osmo_stderr_target, 0);
 	bssgp_set_log_ss(DBSSGP);
-	log_parse_category_mask(osmo_stderr_target, "DL1IF,1:DRLCMAC,1:DRLCMACDATA,1:DRLCMACDL,1:"
-				"DRLCMACUL,1:DRLCMACSCHED,1:DRLCMACMEAS,3:DNS,3:DBSSGP,3:DPCU,5");
+	log_parse_category_mask(osmo_stderr_target, "DRLCMAC,1:DRLCMACDATA,3:DRLCMACDL,3:DRLCMACUL,3:"
+				"DRLCMACSCHED,1:DRLCMACMEAS,3:DNS,3:DBSSGP,3:DPCU,5:"
+				"DL1IF,6:DTBF,1:DTBFUL,1:DTBFDL,1:");
 
 	vty_init(&pcu_vty_info);
 	pcu_vty_init(&gprs_log_info);
